"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import pyomo.environ as pyomo

from pycity_scheduling.util import extract_pyomo_values


class OptimizationEntity(object):
    """
    Base class for all optimization entities.

    This class provides functionality common to all entities which take part
    in the scheduling optimization.

    Parameters
    ----------
    environment : Environment
        Common Environment instance.
    """

    static_entity_id = 0

    def __init__(self, environment, *args, **kwargs):
        OptimizationEntity.static_entity_id += 1
        self.id = OptimizationEntity.static_entity_id
        self._id_string = "{0:05d}".format(self.id)
        self._long_id = ""
        self._kind = ""

        self.objective = None
        self.timer = environment.timer
        self.schedules = {'default': {}, 'ref': {}}

        self._var_order = []
        self._var_funcs = {}
        self.current_schedule_active = 'default'
        self.model = None

        if hasattr(super(), "__module__"):
            # This allows ElectricalEntity and ThermalEntity to be instantiated on their own:
            super().__init__(environment, *args, **kwargs)
        else:
            super().__init__()

    def __str__(self):
        return self._long_id

    def __repr__(self):
        return ("<OptimizationEntity of kind " + self._kind
                + " with ID: " + self._long_id + ">")

    @property
    def op_horizon(self):
        """
        Number of time steps in a scheduling period.
        """
        return self.timer.timesteps_used_horizon

    @property
    def op_time_vec(self):
        """
        Iterator over scheduling period.
        """
        return range(self.timer.timesteps_used_horizon)

    @property
    def simu_horizon(self):
        """
        Number of time steps in the whole simulation horizon.
        """
        return self.timer.simu_horizon

    @property
    def time_slot(self):
        """
        Length of a time step as a portion of an hour.

        Examples
        --------
        time step length = 60 mins => time_slot = 1
        time step length = 15 mins => time_slot = 0.25
        """
        return self.timer.time_slot

    @property
    def timestep(self):
        """
        Time step indicating the current scheduling.
        """
        return self.timer.current_timestep

    @property
    def op_slice(self):
        """
        Slice to select values of current scheduling from whole horizon.
        """
        t1 = self.timer.current_timestep
        t2 = t1 + self.timer.timesteps_used_horizon
        return slice(t1, t2)

    def populate_model(self, model, mode=""):
        """
        Add entity block to pyomo ConcreteModel.

        Places the block with the name of the entity in the ConcreteModel.

        Parameters
        ----------
        model : pyomo.ConcreteModel
        mode : str, optional
            Specifies which set of constraints to use.

            - `convex`  : Use linear constraints
            - `integer`  : May use integer variables
        """
        # generate empty pyomo block
        self.model = pyomo.Block()
        setattr(model, "_".join([self._kind, self._id_string]), self.model)
        # add time
        self.model.t = pyomo.RangeSet(0, self.op_horizon-1)
        return

    def update_model(self, mode=""):
        """
        Update block parameters and bounds.

        Set parameters and bounds according to the current situation of the
        device according to the previous schedule and the current forecasts.

        Parameters
        ----------
        mode : str, optional
            Specifies which set of constraints to use.

            - `convex`  : Use linear constraints
            - `integer`  : Use integer variables representing discrete control decisions
        """
        pass

    def update_schedule(self, verbose=False):
        """
        Update the schedule with the scheduling model solution.

        Retrieve the solution from the scheduling model and write it to the
        schedule. The model must be optimal. The time / position of the
        solution in the schedule is determined by `self.timer.current_timestep`.

        Parameters
        ----------
        verbose : bool, optional
            Verbose mode and output.
        """
        op_slice = self.op_slice
        for name in self._var_order:
            schedule = self.schedule[name]
            if name in self._var_funcs and not hasattr(self.model, name + "_vars"):
                func = self._var_funcs[name]
                if func. __code__.co_argcount == 1:
                    schedule[op_slice] = func(self.model)
                if func. __code__.co_argcount == 2:
                    schedule[op_slice] = func(self, self.model)
            else:
                if hasattr(self, "model") and hasattr(self.model, name + "_vars"):
                    values = extract_pyomo_values(getattr(self.model, name + "_vars"), verbose=verbose)
                    if len(values) < self.op_horizon:
                        values = np.pad(values, (0, len(schedule[op_slice]) - len(values)), mode='constant')
                    schedule[op_slice] = values
        return

    def set_objective(self, objective):
        """
        Set a new objective to be returned by get_objective.

        Parameters
        ----------
        objective : str
            Objective for the scheduling.

            - 'none' : No objective (leave all flexibility to other participants).
        """
        self.objective = objective
        return

    def get_objective(self, coeff=1):
        """
        Objective function for entity level scheduling.

        Return the objective function of the entity weighted with
        coeff.

        Parameters
        ----------
        coeff : float, optional
            Coefficient for the objective function.

        Returns
        -------
        ExpressionBase :
            Objective function.
        """
        if self.objective is None or self.objective == 'none':
            return 0
        else:
            raise ValueError(
                "Objective {} is not implemented by entity {}.".format(self.objective, self.__class__.__name__)
            )

    def get_decision_var(self):
        """
        Return the decision varaible(s) of the entity's objective function.

        Parameters
        ----------

        Returns
        -------
        pyomo.Var :
            Decision variable.
        """
        if self.objective is None or self.objective == 'none':
            m = self.model
            m.pseudo_var = pyomo.Var(m.t, domain=pyomo.Reals, initialize=0)
            return m.pseudo_var

    def reset(self, schedule=None):
        """
        Reset all values of specified schedule.

        Parameters
        ----------
        schedule : str, optional
            Specify which schedule to reset.

            - `None` : Resets all schedules
            - 'default' : Resets normal schedule
            - 'ref' : Resets reference schedule
        """
        if schedule is None:
            for name in self.schedules.keys():
                self.new_schedule(name)
        else:
            self.new_schedule(schedule)
        return

    def get_entities(self):
        """
        Yield all lowest contained entities.

        Yields
        ------
        Lowest contained entities or `self`.
        """
        top = False
        for entity in self.get_lower_entities():
            top = True
            yield from entity.get_entities()
        if not top:
            yield self

    def get_lower_entities(self):
        """
        Yield all lower-level entities.

        Yields
        ------
        All contained entities.
        """
        return iter(())

    def get_all_entities(self):
        """
        Yield all entities.

        Yields
        ------
        All contained entities and sub-entities.
        """
        yield self
        for entity in self.get_lower_entities():
            yield from entity.get_all_entities()

    def new_var(self, name, dtype=np.float64, func=None):
        """
        Create a new entry and empty schedule for variable with specified name.

        Parameters
        ----------
        name : str
            Name to access new variable with.
        dtype : numpy.dtype, optional
            Data type which should be used for new schedule.
        func : Callable[[int], Any], optional
            Function to generate schedule with.
            If `None`, schedule is generated with values of variables.
        """
        if func is not None:
            self._var_funcs[name] = func
        for schedule in self.schedules.values():
            schedule[name] = np.full(self.timer.simu_horizon, 0, dtype=dtype)
        self._var_order.append(name)
        return

    def new_schedule(self, schedule):
        """
        Create a new schedule with default values.

        Parameters
        ----------
        schedule : str
            Name of new schedule.
        """
        self.schedules[schedule] = {name: np.full_like(entries, 0) for name, entries in self.schedule.items()}
        for e in self.get_lower_entities():
            e.new_schedule(schedule)
        return

    def copy_schedule(self, dst=None, src=None, name=None):
        """
        Copy values of one schedule in another schedule.

        Parameters
        ----------
        dst : str
            Name of schedule to insert values into.
            If `None`, use current schedule.
        src : str
            Name of schedule to copy values from.
            If `None`, use current schedule.
        name : str
            Name of variable to copy sub schedule of.
            If `None`, copy all variables between schedules.
        """
        if dst is None:
            dst = self.current_schedule_active
        elif src is None:
            src = self.current_schedule_active
        src_schedule = self.schedules[src]
        if name is None:
            self.schedules[dst] = {key: entries.copy() for key, entries in src_schedule.items()}
        else:
            if dst not in self.schedules:
                self.new_schedule(dst)
            self.schedules[dst][name] = src_schedule[name].copy()
        for e in self.get_lower_entities():
            e.copy_schedule(dst, src, name)
        return

    def load_schedule(self, schedule):
        """
        Copy values of one schedule in another schedule.

        Parameters
        ----------
        schedule : str
        Name of schedule to set as current schedule.
        """
        self.current_schedule_active = schedule
        for e in self.get_lower_entities():
            e.load_schedule(schedule)
        return

    def load_schedule_into_model(self, schedule=None, nudge=True):
        """
        Overwrites the values in the entity model with the values in the schedule.

        Parameters
        ----------
        schedule : str
            Name of schedule to load values from.
            If `None`, use current schedule.
        nudge : bool, optional
            If a variable (slightly) exceeds the lower or upper bound, then use the bound as the actual schedule value.
        """
        if schedule is None:
            s = self.schedule
        else:
            s = self.schedules[schedule]
        op_slice = self.op_slice
        for var_name, var_schedule in s.items():
            var_schedule = var_schedule[op_slice]
            if not hasattr(self.model, var_name + "_vars"):
                continue
            vars = getattr(self.model, var_name + "_vars")
            for t in range(self.op_horizon):
                if t in vars:
                    value = var_schedule[t]
                    if nudge:
                        if vars[t].lb is not None and value < vars[t].lb and np.isclose(vars[t].lb, value):
                            value = vars[t].lb
                        elif vars[t].ub is not None and value > vars[t].ub and np.isclose(vars[t].ub, value):
                            value = vars[t].ub
                    vars[t].value = value

        for e in self.get_lower_entities():
            e.load_schedule_into_model(schedule)
        return

    @property
    def schedule(self):
        """
        The current loaded schedule.
        """
        return self.schedules[self.current_schedule_active]

    def __getattr__(self, item):
        if isinstance(item, str):
            items = item.split("_")
            if len(items) >= 2:
                if items[-1] == "schedule":
                    if items[-2] in self.schedules and "_".join(items[:-2]) in self.schedule:
                        schedule = self.schedules[items[-2]]
                        varname = "_".join(items[:-2])
                    else:
                        schedule = self.schedule
                        varname = "_".join(items[:-1])
                    schedule = schedule.get(varname, None)
                    if schedule is not None:
                        return schedule
        raise AttributeError(item)

    def __setattr__(self, attr, value):
        if isinstance(attr, str):
            attrs = attr.split("_")
            if len(attrs) >= 2:
                if attrs[-1] == "schedule":
                    if attrs[-2] in self.schedules and "_".join(attrs[:-2]) in self.schedule:
                        schedule = self.schedules[attrs[-2]]
                        varname = "_".join(attrs[:-2])
                    else:
                        schedule = self.schedule
                        varname = "_".join(attrs[:-1])
                    schedule[varname] = value
                    return
        super().__setattr__(attr, value)



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import pyomo.environ as pyomo
from warnings import warn
import pycity_base.classes.demand.electrical_demand as ed

from pycity_scheduling.classes.electrical_entity import ElectricalEntity
from pycity_scheduling import util
from pycity_scheduling import constants


class DeferrableLoad(ElectricalEntity, ed.ElectricalDemand):
    """
    Extension of pyCity_base class ElectricalDemand for scheduling purposes.

    The Load will always run once in the op_horizon

    Parameters
    ----------
    environment : Environment
        Common Environment instance.
    p_el_nom : float
        Nominal electric power in [kW].
    e_consumption : float
         Power to be consumed over the op_horizon in [kWh].
    load_time : array of binaries, optional
        Indicator when deferrable load can be turned on. Defaults to always.
        `load_time[t] == 0`: device is off in t
        `load_time[t] == 1`: device can be turned on in t
        It must contain at least one `0` otherwise the model will become
        infeasible. Its length has to be consistent with `lt_pattern`.
    lt_pattern : str, optional
        Define how the `load_time` profile is to be used

        - `None` : Profile matches simulation horizon.
        - 'daily' : Profile matches one day.
        - 'weekly' : Profile matches one week.

    Raises
    ------
    ValueError :
        If `lt_pattern` does not match `load_time`.

    Notes
    -----
    - DLs offer sets of constraints for operation. In the `convex` mode the following
      constraints and bounds are generated by the DL:

    .. math::
        p_{el\\_nom} \\geq p_{el\\_i} \\geq 0, & \\quad \\text{if} \\quad lt\\_pattern_i = 1  \\\\
        p_{el\\_i} = 0, & \\quad \\text{else}

    .. math::
        \\sum_i p_{el\\_i} * \\Delta t = e_{consumption}

    - The constraints are replaced in integer mode with the following constraints:

    .. math::
        \\sum_i p_{state\\_i} &=& 1 \\\\
        runtime &=& \\lfloor \\frac{p_{el\\_nom} * \\Delta t}{e_{consumption}} \\rceil \\\\
        p_{el\\_i} &=& p_{el\\_nom} * \\sum_{j=i-runtime+1}^{i} p_{state\\_j} \\\\
        p_{state\\_i} &=& 0, \\quad \\text{if} \\quad \\sum_{j=i}^{i+runtime-1}
        lt\\_pattern_j = runtime \\\\

    - These constraints do not take the previous values before the current optimization
      horizon into account. In the optimization horizon :math:`e_{consumption}` always
      has to be consumed.
    """

    def __init__(self, environment, p_el_nom, e_consumption, load_time=None, lt_pattern=None):
        shape = environment.timer.timesteps_total
        super().__init__(environment, 0, np.zeros(shape))

        self._long_id = "DL_" + self._id_string

        self.p_el_nom = p_el_nom
        self.e_consumption = e_consumption
        if load_time is None:
            lt_pattern = None
            load_time = np.ones(self.simu_horizon)
        self.load_time = util.compute_profile(self.timer, load_time, lt_pattern)

        if len(load_time) != self.op_horizon:
            warn(
                ("The DeferrableLoad {} will always run once in the op_horizon and not once in the lt_pattern.\n" +
                 "If a different behaviour is intended, initializing multiple DeferrableLoads should be considered.")
                .format(self._long_id)
            )

        self.new_var("p_start", dtype=np.bool_, func=DeferrableLoad._get_start)

        self.runtime = int(round(self.e_consumption / (self.p_el_nom * self.time_slot)))

        if self.runtime == 0:
            raise ValueError(
                ("The DeferrableLoad {} has a runtime of zero timesteps, which leads to infeasible constraints. "
                 "Consider removing it.")
                .format(self._long_id)
            )

    def _get_start(self, model):
        cumsum = np.zeros((self.op_horizon+1))
        np.cumsum(self.schedule["p_el"][self.op_slice], out=cumsum[1:])
        runtime_consumptions = cumsum[self.runtime:] - cumsum[:-self.runtime]
        starts = np.zeros(self.op_horizon, dtype=np.bool_)
        starts[np.argmax(runtime_consumptions)] = True
        return starts

    def populate_model(self, model, mode="convex"):
        """
        Add device block to pyomo ConcreteModel

        Call parent's `populate_model` method and set the upper bounds to the
        nominal power or zero depending on `self.load_time`. Also set a constraint
        for the minimum load. If mode == `integer` add binary variables to model
        load as one block that can be shifted in time.

        Parameters
        ----------
        model : pyomo.ConcreteModel
        mode : str, optional
            Specifies which set of constraints to use.

            - `convex`  : Use linear constraints
            - `integer`  : Uses integer variables to restrict the DL to operate
               at nominal load or no load and restricts the DL to consume
               E_Min_Consumption when DL is started without returning to a no
               load state

        """
        super().populate_model(model, mode)
        m = self.model

        def p_el_vars_ub_rule(model, t):
            return model.p_el_vars[t] <= constants.INFINITY
        m.p_el_vars_ub_constr = pyomo.Constraint(m.t, rule=p_el_vars_ub_rule)

        if mode == "convex":
            if self.e_consumption > self.op_horizon * self.time_slot * self.p_el_nom:
                warn(
                    ("DeferrableLoad {} is not able to consume enough power in the given op_horizon," +
                     "which will render the model infeasible")
                    .format(self._long_id)
                )

            # consume e_consumption the op_horizon
            def p_consumption_rule(model):
                return sum(model.p_el_vars[t] for t in range(len(model.p_el_vars))) * self.time_slot == \
                       self.e_consumption
            m.p_consumption_constr = pyomo.Constraint(rule=p_consumption_rule)

        elif mode == "integer":
            self.runtime = self.e_consumption / (self.p_el_nom * self.time_slot)
            rounded_runtime = int(round(self.runtime))
            if not np.isclose(self.runtime, rounded_runtime):
                warn("Consumption of DLs is in integer mode always a multiple of " +
                     "p_el_nom * dt, which is larger than e_consumption.")
            self.runtime = rounded_runtime

            if self.runtime > self.op_horizon:
                raise ValueError(
                    ("DeferrableLoad {} is not able to complete its operation in the given op_horizon, " +
                     "which will render the model infeasible.")
                    .format(self._long_id)
                )

            # create binary variables representing if operation begins in timeslot t
            # Since the DL has to finish operation when the op_horizon ends, binary variables representing a too late
            # start can be omitted.
            m.p_start_vars = pyomo.Var(pyomo.RangeSet(0, self.op_horizon-self.runtime), domain=pyomo.Binary)

            def p_start_vars_ub_rule(model, t):
                return model.p_start_vars[t] <= 1.0
            m.p_start_vars_ub_constr = pyomo.Constraint(pyomo.RangeSet(0, self.op_horizon-self.runtime),
                                                        rule=p_start_vars_ub_rule)

            # coupling the start variable to the electrical variables following
            def state_coupl_rule(model, t):
                return model.p_el_vars[t] == self.p_el_nom * sum(
                       (model.p_start_vars[tau] for tau in range(max(0, t+1-self.runtime),
                                                                 min(self.op_horizon-self.runtime+1, t+1))))
            m.state_coupl_integer_constr = pyomo.Constraint(m.t, rule=state_coupl_rule)

            # run once in the op_horizon
            def run_once_rule(model):
                if len(model.p_start_vars) > 0:
                    return sum(model.p_start_vars[t] for t in range(len(model.p_start_vars))) == 1.0
                else:
                    return pyomo.sum_product(model.p_start_vars) == 1.0
            m.run_once_integer_constr = pyomo.Constraint(rule=run_once_rule)

        else:
            raise ValueError(
                "Mode %s is not implemented by class DeferrableLoad." % str(mode)
            )
        return

    def update_model(self, mode="convex"):
        m = self.model
        if mode == "convex":
            load_time = self.load_time[self.op_slice]

            p_el_vars_ub_constr_active = m.p_el_vars_ub_constr.active

            def p_el_vars_ub_rule(model, t):
                if t in self.op_time_vec:
                    if load_time[t] == 1:
                        return model.p_el_vars[t] <= self.p_el_nom
                    else:
                        return model.p_el_vars[t] <= 0.0
                else:
                    return model.p_el_vars[t] <= constants.INFINITY
            try:
                m.del_component(m.p_el_vars_ub_constr)
            except AttributeError:
                pass
            m.p_el_vars_ub_constr = pyomo.Constraint(m.t, rule=p_el_vars_ub_rule)

            if p_el_vars_ub_constr_active is False:
                m.p_el_vars_ub_constr.deactivate()

        elif mode == "integer":
            if (self.timestep + self.op_horizon < self.simu_horizon and
                    self.load_time[self.timestep + self.op_horizon - 1] == 1 and
                    self.load_time[self.timestep + self.op_horizon] == 1):
                warn(
                    ("DeferrableLoad {} will not consider a start which would result in power consumption" +
                     "outside the op_horizon")
                    .format(self._long_id)
                     )

            load_time = self.load_time[self.op_slice]

            p_start_vars_ub_value = np.zeros(self.op_horizon-self.runtime+1)
            for t, start_var in m.p_start_vars.items():
                if all(lt == 1 for lt in load_time[t:t+self.runtime]):
                    p_start_vars_ub_value[t] = 1.0

            p_start_vars_ub_constr_active = m.p_start_vars_ub_constr.active

            def p_start_vars_ub_rule(model, t):
                return model.p_start_vars[t] <= p_start_vars_ub_value[t]
            try:
                m.del_component(m.p_start_vars_ub_constr)
            except AttributeError:
                pass
            m.p_start_vars_ub_constr = pyomo.Constraint(pyomo.RangeSet(0, self.op_horizon-self.runtime),
                                                        rule=p_start_vars_ub_rule)

            if p_start_vars_ub_constr_active is False:
                m.p_start_vars_ub_constr.deactivate()

        else:
            raise ValueError(
                "Mode %s is not implemented by DL." % str(mode)
            )
        return

    def get_objective(self, coeff=1):
        """
        Objective function for entity level scheduling.

        Return the objective function of the deferrable load weighted with
        coeff. Quadratic term minimizing the deviation from the optimal
        loadcurve.

        Parameters
        ----------
        coeff : float, optional
            Coefficient for the objective function.

        Returns
        -------
        ExpressionBase :
            Objective function.
        """
        m = self.model
        max_loading_time = sum(self.load_time) * self.time_slot
        optimal_p_el = self.e_consumption / max_loading_time
        obj = 0
        for t in range(self.op_horizon):
            obj += coeff * m.p_el_vars[t] * m.p_el_vars[t]
        obj -= 2 * coeff * optimal_p_el * sum(m.p_el_vars[t] for t in range(self.op_horizon))
        return obj

    def get_decision_var(self):
        return self.model.p_el_vars



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import pyomo.environ as pyomo

from pycity_scheduling.classes.optimization_entity import OptimizationEntity
from pycity_scheduling import constants


class ElectricalEntity(OptimizationEntity):
    """
    Base class for all electrical entities derived from OptimizationEntity.

    This class provides functionality common to all electrical entities.
    It adds variables for the electrical demand / supply of the entity to the
    block.

    Parameters
    ----------
    model : pyomo.ConcreteModel
    mode : str, optional
        Specifies which set of constraints to use.

        - `convex`  : Use linear constraints
        - `integer`  : Use same constraints as convex mode

    Notes
    -----
    - EEs add the :math:`p_{el}` variable to the model. When not modified
      by other classes, the following constraint is added:

    .. math::
        p_{el} \\geq 0
    """

    def __init__(self, environment, *args, **kwargs):
        super().__init__(environment, *args, **kwargs)

        self.new_var("p_el")

    def populate_model(self, model, mode="convex"):
        super().populate_model(model, mode)
        m = self.model

        if mode in ["convex", "integer"]:
            m.p_el_vars = pyomo.Var(m.t, domain=pyomo.Reals, bounds=(-constants.INFINITY, constants.INFINITY),
                                    initialize=0)

            def p_el_vars_domain_lb_rule(model, t):
                return model.p_el_vars[t] >= -constants.INFINITY
            m.p_el_vars_domain_lb_constr = pyomo.Constraint(m.t, rule=p_el_vars_domain_lb_rule)

            def p_el_vars_domain_ub_rule(model, t):
                return model.p_el_vars[t] <= constants.INFINITY
            m.p_el_vars_domain_ub_constr = pyomo.Constraint(m.t, rule=p_el_vars_domain_ub_rule)

            if self.objective == "max-consumption":
                m.max_consumption_var = pyomo.Var(domain=pyomo.Reals, initialize=0)

                def max_consumption_lb_rule(model):
                    return model.max_consumption_var >= 0.0
                m.max_consumption_lb_constr = pyomo.Constraint(rule=max_consumption_lb_rule)

                def p_consumption_rule(model, t):
                    return model.max_consumption_var >= m.p_el_vars[t]
                m.p_cons_constr = pyomo.Constraint(m.t, rule=p_consumption_rule)

                def p_generation_rule(model, t):
                    return model.max_consumption_var >= -m.p_el_vars[t]
                m.p_gen_constr = pyomo.Constraint(m.t, rule=p_generation_rule)

            if self.objective == "self-consumption" or self.objective == "co2":
                m.p_export_vars = pyomo.Var(m.t, domain=pyomo.Reals, initialize=0)
                m.p_import_vars = pyomo.Var(m.t, domain=pyomo.Reals, initialize=0)

                def p_export_vars_lb_rule(model, t):
                    return model.p_export_vars[t] >= 0.0
                m.p_export_vars_lb_constr = pyomo.Constraint(m.t, rule=p_export_vars_lb_rule)

                def p_import_vars_lb_rule(model, t):
                    return model.p_import_vars[t] >= 0.0
                m.p_import_vars_lb_constr = pyomo.Constraint(m.t, rule=p_import_vars_lb_rule)

                def p_self_consumption_rule(model, t):
                    return model.p_import_vars[t] - model.p_export_vars[t] == m.p_el_vars[t]
                m.p_self_cons_constr = pyomo.Constraint(m.t, rule=p_self_consumption_rule)

            if self.objective == "flexibility-quantification":
                m.max_p_flex_vars = pyomo.Var(m.t, domain=pyomo.Reals, initialize=0)

                def max_p_flex_vars_lb_rule(model, t):
                    return model.max_p_flex_vars_vars[t] >= 0.0
                m.max_p_flex_vars_lb_constr = pyomo.Constraint(m.t, rule=max_p_flex_vars_lb_rule)

                def max_p_flex_consumption_rule(model, t):
                    return model.max_p_flex_vars[t] >= m.p_el_vars[t]
                m.max_p_flex_cons_constr = pyomo.Constraint(m.t, rule=max_p_flex_consumption_rule)

                def max_p_flex_generation_rule(model, t):
                    return model.max_p_flex_vars[t] >= -m.p_el_vars[t]
                m.max_p_flex_gen_constr = pyomo.Constraint(m.t, rule=max_p_flex_generation_rule)
        else:
            raise ValueError("Mode %s is not implemented by electrical entity." % str(mode))
        return

    def get_objective(self, coeff=1):
        if self.objective == 'peak-shaving':
            obj = 0
            for t in range(self.op_horizon):
                obj += self.model.p_el_vars[t] * self.model.p_el_vars[t]
            return coeff * obj
        if self.objective == 'price':
            prices = self.environment.prices.tou_prices
            prices = prices[self.op_slice]
            s = sum(abs(prices))
            if s > 0:
                prices = prices * self.op_horizon / s
                obj = 0
                for t in range(self.op_horizon):
                    obj += prices[t] * self.model.p_el_vars[t]
                return coeff * obj
            else:
                return 0
        if self.objective == 'co2':
            prices = self.environment.prices.co2_prices
            prices = prices[self.op_slice]
            s = sum(abs(prices))
            if s > 0:
                prices = prices * self.op_horizon / s
                obj = 0
                for t in range(self.op_horizon):
                    obj += prices[t] * self.model.p_import_vars[t]
                return coeff * obj
            else:
                return 0
        if self.objective == "max-consumption":
            if not hasattr(self.model, "max_consumption_var"):
                raise ValueError("Objective 'max-consumption' needs to be selected during populate_model call.")
            if coeff < 0:
                raise ValueError("Setting a coefficient below zero is not supported for the max-consumption objective.")
            return coeff * self.model.max_consumption_var * self.model.max_consumption_var
        if self.objective == "self-consumption":
            if not hasattr(self.model, "p_export_vars") or not hasattr(self.model, "p_import_vars"):
                raise ValueError("'Objective self-consumption' needs to be selected during populate_model call.")
            if coeff < 0:
                raise ValueError("Setting a coefficient below zero is not supported for the self-consumption "
                                 "objective.")
            obj = 0
            for t in range(self.op_horizon):
                obj += self.model.p_export_vars[t] * self.model.p_export_vars[t]
            return coeff * obj
        if self.objective == 'least-squares':
            obj = 0
            for t in range(self.op_horizon):
                obj += (self.model.p_el_vars[t] - self.least_squares_profile[t]) * \
                       (self.model.p_el_vars[t] - self.least_squares_profile[t])
            return coeff * obj
        if self.objective == 'flexibility-quantification':
            if not hasattr(self.model, "max_p_flex_vars"):
                raise ValueError("Objective 'flexibility-quantification' needs to be selected during populate_model "
                                 "call.")
            obj = 0
            for t in range(self.op_horizon):
                obj += self.model.max_p_flex_vars[t] * self.model.max_p_flex_vars[t]
            return coeff * obj
        return super().get_objective(coeff)

    def get_decision_var(self):
        if self.objective in ['price', 'co2', 'peak-shaving', 'least-squares']:
            return self.model.p_el_vars
        if self.objective == "max-consumption":
            return self.model.max_consumption_var
        if self.objective == "self-consumption":
            return self.model.p_export_var
        if self.objective == 'flexibility-quantification':
            return self.model.max_p_flex_var
        return super().get_decision_var()



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import unittest
import os
import importlib.util


class TestExamples(unittest.TestCase):
    def setUp(self):
        this_dir = os.path.dirname(__file__)
        self.example_dir = os.path.join(this_dir, "../../examples")
        self.files = os.listdir(self.example_dir)
        if len(self.files) == 0:
            self.skipTest("No example files found.")
        return

    def test_all_examples(self):
        for file in sorted(self.files):
            filepath = os.path.join(self.example_dir, file)
            example_name, file_ext = os.path.splitext(os.path.split(filepath)[-1])
            if file_ext.lower() == '.py' and example_name != '__init__':
                spec = importlib.util.spec_from_file_location(example_name, filepath)
                example_module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(example_module)
                example_module.main(do_plot=False)
        return



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np

import pycity_scheduling.classes as classes
import pycity_scheduling.algorithms as algs
from pycity_scheduling.util import factory
from pycity_scheduling.util.metric import calculate_costs, peak_to_average_ratio, peak_reduction_ratio, \
                                          self_consumption, autarky, absolute_flexibility_gain


# This is a simple power scheduling example using the central optimization algorithm in order to demonstrate the
# capabilities of the pycity_scheduling framework when it comes to the evaluation of metrics. The evaluation of metrics
# is usually an important post-processing step.


def main(do_plot=False):
    print("\n\n------ Example 30: Post-Processing Metrics Evaluation ------\n\n")

    # Use a standard environment of 24 hours with hourly resolution (=60min=3600sec):
    env = factory.generate_standard_environment(step_size=3600, op_horizon=24)

    # Make it "attractive" for the consumers to shift demand into the second half of the scheduling period:
    env.prices.tou_prices = np.array([20]*12 + [10]*12)

    # City district / district operator objective is set to peak-shaving:
    cd = classes.CityDistrict(env, objective='peak-shaving')

    # The sample building in this example comes with a constant electrical and space heating load of 10kW,
    # thermal storage of capacity 20kWh and an electric heater of thermal nominal power 20kW:
    bd = factory.generate_simple_building(env, fl=10, sh=10, ths=20, eh=20)
    cd.addEntity(bd, (0, 0))

    # Perform a 'pseudo' stand-alone power scheduling, where each device is scheduled on its own (=no coordination):
    o1 = algs.StandAlone(cd)
    o1.solve()

    # Results are now in the _Ref schedules:
    bd.copy_schedule("ref")
    cd.copy_schedule("ref")

    # Now perform a coordinated scheduling with district operator and customer objectives:
    o2 = algs.CentralOptimization(cd)
    o2.solve()

    # Evaluate and print different metrics:
    np.set_printoptions(formatter={'float': '{: >8.3f}'.format})
    print('Comparing stand-alone with optimized case:')
    print('Building p_el:')
    print(bd.p_el_ref_schedule)
    print(bd.p_el_schedule)
    print('Optimized costs:        {: >8.3f}'.format(calculate_costs(bd)))
    bd.load_schedule("ref")
    print('Stand-alone costs:      {: >8.3f}'
          .format(calculate_costs(bd)))
    bd.load_schedule("default")
    print('Optimized PAR:          {: >8.3f}'.format(peak_to_average_ratio(bd)))
    bd.load_schedule("ref")
    print('Stand-alone PAR:        {: >8.3f}'
          .format(peak_to_average_ratio(bd)))
    bd.load_schedule("default")
    print('PRR:                    {: >8.3f}'.format(peak_reduction_ratio(bd, "ref")))
    print('Self-consumption ratio: {: >8.3f}'.format(self_consumption(bd)))
    print('Autarky ratio:          {: >8.3f}'.format(autarky(bd)))
    print('Absolute flex. gain:    {: >8.3f}'.format(absolute_flexibility_gain(cd, "ref")))
    return


# Conclusion:
# In contrast to the uncoordinated case (stand-alone algorithm) both the total costs as well as the power peaks are
# (slightly) reduced in the coordinated case in this exemplary scheduling case. This is because of the price and
# peak-shaving objectives of the buildings and the city district. In general, the evaluation of different metrics
# constitutes a helpful step when it comes to the comparison of different city district setups and/or power scheduling
# algorithms.


if __name__ == '__main__':
    # Run example:
    main(do_plot=True)



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import pyomo.environ as pyomo
import pycity_base.classes.supply.photovoltaic as pv

from pycity_scheduling.classes.electrical_entity import ElectricalEntity


class Photovoltaic(ElectricalEntity, pv.PV):
    """
    Extension of pyCity_base class PV for scheduling purposes.

    Parameters
    ----------
    environment : Environment
        Common Environment instance.
    method : int
        - 0 : Calculate PV power based on an area in m^2 equipped with PV panels
        - 1 : Calculate PV power based on the installed PV peak power in kWp
    area : float, optional
        PV unit installation area in m^2 for `method=0`.
    peak_power : float, optional
        PV peak power installation in kWp for `method=1`.
    eta_noct : float, optional
        Electrical efficiency at NOCT conditions (without unit) for `method=0`.
        NOCT conditions: See manufacturer's data sheets or
        Duffie, Beckman - Solar Engineering of Thermal Processes (4th ed.), page 759
    radiation_noct : float, optional
        Nominal solar radiation at NOCT conditions (in W/m^2)
        NOCT conditions: See manufacturer's data sheets or
        Duffie, Beckman - Solar Engineering of Thermal Processes (4th ed.), page 759
    t_cell_noct : float, optional
        Nominal cell temperature at NOCT conditions (in degree Celsius)
        NOCT conditions: See manufacturer's data sheets or
        Duffie, Beckman - Solar Engineering of Thermal Processes (4th ed.), page 759
    t_ambient_noct : float, optional
        Nominal ambient air temperature at NOCT conditions (in degree Celsius)
        NOCT conditions: See manufacturer's data sheets or
        Duffie, Beckman - Solar Engineering of Thermal Processes (4th ed.), page 759
    alpha_noct : float, optional
        Temperature coefficient at NOCT conditions (without unit)
        NOCT conditions: See manufacturer's data sheets or
        Duffie, Beckman - Solar Engineering of Thermal Processes (4th ed.), page 759
    beta : float, optional
        Slope, the angle (in degree) between the plane of the surface in
        question and the horizontal. 0 <= beta <= 180. If beta > 90, the
        surface faces downwards.
    gamma : float, optional
        Surface azimuth angle. The deviation of the projection on a
        horizontal plane of the normal to the surface from the local
        meridian, with zero due south, east negative, and west positive.
        -180 <= gamma <= 180
    tau_alpha : float, optional
        Optical properties of the PV unit. Product of absorption and
        transmission coeffients.
        According to Duffie, Beckman - Solar Engineering of Thermal
        Processes (4th ed.), page 758, this value is typically close to 0.9
    force_renewables : bool, optional
        `True` if generation may not be reduced for optimization purposes.

    Notes
    -----
    - The following constraint is added for removing the bounds from EE:

    .. math::
        p_{el} &=& -p_{el\\_supply}, & \\quad \\text{if force_renewables} \\\\
        0 \\geq p_{el} &\\geq& -p_{el\\_supply} , & \\quad \\text{else}
    """

    def __init__(self, environment, method, area=0.0, peak_power=0.0, eta_noct=0.18, radiation_noct=1000.0,
                 t_cell_noct=45.0, t_ambient_noct=20.0, alpha_noct=0, beta=0, gamma=0, tau_alpha=0.9,
                 force_renewables=True):
        super().__init__(environment, method, area, peak_power, eta_noct, radiation_noct, t_cell_noct, t_ambient_noct,
                         alpha_noct, beta, gamma, tau_alpha)
        self._long_id = "PV_" + self._id_string

        self.force_renewables = force_renewables
        self.getPower(currentValues=False)
        ts = self.timer.time_in_year(from_init=True)
        self.p_el_supply = self.total_power[ts:ts+self.simu_horizon] / 1000

    def populate_model(self, model, mode="convex"):
        super().populate_model(model, mode)
        m = self.model
        timestep = self.timestep

        def p_el_vars_lb_rule(model, t):
            return model.p_el_vars[t] >= -self.p_el_supply[timestep + t]
        m.p_el_vars_lb_constr = pyomo.Constraint(m.t, rule=p_el_vars_lb_rule)

        def p_el_vars_ub_rule(model, t):
            if self.force_renewables:
                return model.p_el_vars[t] <= -self.p_el_supply[timestep + t]
            else:
                return model.p_el_vars[t] <= 0.0
        m.p_el_vars_ub_constr = pyomo.Constraint(m.t, rule=p_el_vars_ub_rule)

        return

    def update_model(self, mode=""):
        m = self.model
        timestep = self.timestep

        p_el_vars_lb_constr_active = m.p_el_vars_lb_constr.active

        def p_el_vars_lb_rule(model, t):
            return model.p_el_vars[t] >= -self.p_el_supply[timestep + t]
        try:
            m.del_component(m.p_el_vars_lb_constr)
        except AttributeError:
            pass
        m.p_el_vars_lb_constr = pyomo.Constraint(m.t, rule=p_el_vars_lb_rule)

        if p_el_vars_lb_constr_active is False:
            m.p_el_vars_lb_constr.deactivate()

        p_el_vars_ub_constr_active = m.p_el_vars_ub_constr.active

        def p_el_vars_ub_rule(model, t):
            if self.force_renewables:
                return model.p_el_vars[t] <= -self.p_el_supply[timestep + t]
            else:
                return model.p_el_vars[t] <= 0.0
        try:
            m.del_component(m.p_el_vars_ub_constr)
        except AttributeError:
            pass
        m.p_el_vars_ub_constr = pyomo.Constraint(m.t, rule=p_el_vars_ub_rule)

        if p_el_vars_ub_constr_active is False:
            m.p_el_vars_ub_constr.deactivate()
        return

    def get_objective(self, coeff=1):
        """
        Objective function of the Photovoltaic.

        Return the objective function of the photovoltaic weighted
        with `coeff`. Depending on `self.force_renewables` leave objective
        function empty or build quadratic objective function to minimize
        discrepancy between available power and produced power.

        Parameters
        ----------
        coeff : float, optional
            Coefficient for the objective function.

        Returns
        -------
        ExpressionBase :
            Objective function.
        """
        m = self.model

        s = 0
        for t in self.op_time_vec:
            s += m.p_el_vars[t] * m.p_el_vars[t]
            s -= 2 * self.p_el_supply[self.op_slice][t] * m.p_el_vars[t]
        return coeff * s

    def get_decision_var(self):
        return self.model.p_el_vars



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import matplotlib.pyplot as plt

import pycity_scheduling.util.factory as factory
import pycity_scheduling.util.debug as debug
from pycity_scheduling.algorithms import *
from pycity_scheduling.classes import *


# In this example, the power schedule for a complex city district scenario is determined. The scenario is built upon the
# district setup as defined in example 'example_17_district_generator.py', but it contains more than 100 buildings and
# is hence considered more complex.

def main(do_plot=False):
    print("\n\n------ Example 20: Scheduling Complex City District ------\n\n")

    # First, create an environment using the factory's "generate_standard_environment" method. The environment
    # automatically encapsulates time, weather, and price data/information.
    env = factory.generate_standard_environment(initial_date=(2018, 12, 6), step_size=900, op_horizon=96)

    # Create 75 single-family houses:
    num_sfh = 75

    # 50% SFH.2002, 30% SFH.2010, 20% SFH.2016 (based on TABULA):
    sfh_distribution = {
        'SFH.2002': 0.5,
        'SFH.2010': 0.3,
        'SFH.2016': 0.2,
    }

    # 50% of the single-family houses are equipped with heat pump, 10% with boiler, and 40% with electric heater:
    sfh_heating_distribution = {
        'HP': 0.5,
        'BL': 0.1,
        'EH': 0.4,
    }

    # All single-family houses are equipped with a fixed load, 20% have a deferrable load, and 30% have an electric
    # vehicle. Moreover, 50% of all single-family houses have a battery unit and 80% have a rooftop photovoltaic unit
    # installation.
    # The values are rounded in case they cannot be perfectly matched to the given number of buildings.
    sfh_device_probs = {
        'FL': 1,
        'DL': 0.2,
        'EV': 0.3,
        'BAT': 0.5,
        'PV': 0.8,
    }

    # Create 25 multi-family houses (number of apartments according to TABULA):
    num_mfh = 25

    # 60% MFH.2002, 20% SFH.2010, 20% SFH.2016 (based on TABULA):
    mfh_distribution = {
        'MFH.2002': 0.6,
        'MFH.2010': 0.2,
        'MFH.2016': 0.2,
    }

    # 40% of the multi-family houses are equipped with heat pump, 20% with boiler, and 40% with electric heater:
    mfh_heating_distribution = {
        'HP': 0.4,
        'BL': 0.2,
        'EH': 0.4,
    }

    # All apartments inside a multi-family houses are equipped with a fixed load, 20% have a deferrable load, and 20%
    # have an electric vehicle. Moreover, 40% of all multi-family houses have a battery unit and 80% have a rooftop
    # photovoltaic unit installation.
    # The values are rounded in case they cannot be perfectly matched to the given number of buildings.
    mfh_device_probs = {
        'FL': 1,
        'DL': 0.2,
        'EV': 0.2,
        'BAT': 0.4,
        'PV': 0.8,
    }

    # Finally, create the desired city district using the factory's "generate_tabula_district" method. The district's
    # district operator's objective is defined as "price" (i.e., day-ahead prices) and the buildings' objectives are
    # also defined as "price" (i.e., time-of-use prices).
    district = factory.generate_tabula_district(env, num_sfh, num_mfh,
                                                sfh_distribution,
                                                sfh_heating_distribution,
                                                sfh_device_probs,
                                                mfh_distribution,
                                                mfh_heating_distribution,
                                                mfh_device_probs,
                                                district_objective='price',
                                                building_objective='price'
                                                )

    # To cover the city district's load, the setup additionally comprises a wind energy converter of approx. 2MWp:
    v = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 99])
    p = np.array([0, 0, 3, 25, 82, 174, 321, 532, 815, 1180, 1580, 1810, 1980, 2050, 2050, 2050, 2050, 2050, 2050, 2050,
                  2050, 2050, 2050, 2050, 2050, 2050, 0, 0])
    wec = WindEnergyConverter(env, velocity=v, power=p, hub_height=78.0)
    district.addEntity(wec, [0, 0])

    # Hierarchically print the district and all buildings/assets:
    debug.print_district(district, 1)

    # Perform the city district scheduling using the central optimization algorithm:
    opt = CentralOptimization(district)
    results = opt.solve()
    district.copy_schedule("district_schedule")

    # Plot the scheduling results:
    plt.plot(district.p_el_schedule)
    plt.ylabel("City District Power [kW]")
    plt.title("Complex City District Scenario - Schedule")
    plt.grid()
    if do_plot:
        plt.show()
    return


# Conclusions:
# The power scheduling for a complex city district scenario can be done easily using pycity_scheduling. Even more than
# 100 (TABULA) buildings could be considered, but this might lead to scalability issues when using the central
# optimization algorithm. Thus, with a growing number buildings, distributed algorithms such as dual decomposition or
# ADMM should be applied.


if __name__ == '__main__':
    # Run example:
    main(do_plot=True)



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import pycity_base.classes.supply.building_energy_system as bes

from pycity_scheduling.classes.entity_container import EntityContainer
from pycity_scheduling.classes.thermal_entity_cooling import ThermalEntityCooling
from pycity_scheduling.classes.thermal_entity_heating import ThermalEntityHeating


class BuildingEnergySystem(EntityContainer, bes.BES):
    """
    Extension of pyCity_base class BES for scheduling purposes.

    Parameters
    ----------
    environment : Environment

    Notes
    -----
    - The constraints generated by a BES are the same as the ones created by an EntityContainer.
    """

    def __init__(self, environment):
        super().__init__(environment)
        self._long_id = "BES_" + self._id_string

    def get_lower_entities(self):
        yield from self.boilers
        yield from self.compression_chillers
        yield from self.chp_units
        yield from self.electrical_heaters
        yield from self.heatpumps
        yield from self.ths_units
        yield from self.tcs_units
        yield from self.battery_units
        yield from self.pv_units

    @property
    def ths_units(self):
        """Provides all THS units."""
        ths_units = []
        for tes in self.tes_units:
            if isinstance(tes, ThermalEntityHeating):
                ths_units.append(tes)
        return ths_units

    @property
    def tcs_units(self):
        """Provides all TCS units."""
        tcs_units = []
        for tes in self.tes_units:
            if isinstance(tes, ThermalEntityCooling):
                tcs_units.append(tes)
        return tcs_units

    def getHasDevices(self,
                      all_devices=True,
                      battery=False,
                      boiler=False,
                      chp=False,
                      chiller=False,
                      electrical_heater=False,
                      heatpump=False,
                      inverter_acdc=False,
                      inverter_dcac=False,
                      pv=False,
                      ths=False,
                      tcs=False):
        """
        Get information if certain devices are installed devices.
        The result is in alphabetical order, starting with "battery"

        Parameters
        ----------
        all_devices : boolean, optional
            If true: Return all installed devices
            If false: Only return the specified devices
        battery : boolean, optional
            Return information about the battery unit
        boiler : boolean, optional
            Return information about the boiler unit
        chp : boolean, optional
            Return information about the chp unit
        chiller : boolean, optional
            Return information about the chiller unit
        electrical_heater : boolean, optional
            Return information about the electric heater unit
        heatpump : boolean, optional
            Return information about the heat pump unit
        inverter_acdc : boolean, optional
            Return information about the AC-DC inverter unit
        inverter_dcac : boolean, optional
            Return information about the DC-AC inverter unit
        pv : boolean, optional
            Return information about the PV unit
        ths : boolean, optional
            Return information about the thermal heating storage unit
        tcs : boolean, optional
            Return information about the thermal cooling storage unit
        """
        result = super().getHasDevices(all_devices=all_devices,
                                       battery=battery,
                                       boiler=boiler,
                                       chp=chp,
                                       compression_chiller=chiller,
                                       electrical_heater=electrical_heater,
                                       heatpump=heatpump,
                                       inverter_acdc=inverter_acdc,
                                       inverter_dcac=inverter_dcac,
                                       pv=pv,
                                       tes=False)
        if all_devices:
            result = list(result)
            result[-1] = (len(self.ths_units) > 0)
            result.append((len(self.tcs_units) > 0))
            result = tuple(result)
        else:
            if ths:
                result += ((len(self.ths_units) > 0),)
            if tcs:
                result += ((len(self.tcs_units) > 0),)
        return result



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import pyomo.core
import pyomo.environ as pyomo

from pycity_scheduling import constants


class Constraint:
    """
    Base class for all generic constraints.

    This class provides functionality common to all generic constraints.
    Generic constraints can be easily added to an entity block.
    """
    def apply(self, model, mode=""):
        """
        Apply constraint to block during populate_model method call.

        Parameters
        ----------
        model : pyomo.Block
            The block corresponding to the entity the constraint should
            be applied to.
        mode : str, optional
            Specifies which set of constraints to use.

            - `convex`  : Use linear constraints
            - `integer`  : May use integer variables
        """
        raise NotImplementedError()


class LowerActivationLimit(Constraint):
    """
    Constraint Class for adding lower activation limits

    This class provides functionality to add lower activation limits
    to entities. Adds no new constraints and variables if not in integer
    mode or if not required. A new state schedule is also created for
    the entity.

    Notes
    -----
    - In `integer` mode the following constraints are added:

    .. math::
        State  \\geq \\frac{var}{var\\_nom} \\geq State * lower\\_activation\\_limit

    """
    def __init__(self, o, var_name, lower_activation_limit, var_nom, min_off_time=1, min_on_time=1):
        """
        Initialize Constraint for a specific entity and create the new
        state schedule.


        Parameters
        ----------
        o : OptimizationEntity
            The entity to add the constraint to.
        var_name : str
            The variable name the constraint should be applied to.
        lower_activation_limit : float (0 <= lowerActivationLimit <= 1)
            Defines the lower activation limit. For example, heat pumps
            are typically able to operate between 50% part load and rated
            load. In this case, lowerActivationLimit would be 0.5
            Two special cases:

            - Linear behaviour: lowerActivationLimit = 0.0
            - Two-point controlled: lowerActivationLimit = 1.0
        var_nom : float
            The maximum or minimum value the variable takes when operating
            under maximum load.
        min_off_time : int, optional
            Minimum number of consecutive time steps the entity must remain turned off, once it has been decided that it
            should turn off. Defaults to '1', i.e., not limited to a consecutive number of time steps.
        min_on_time : int, optional
            Minimum number of consecutive time steps the entity must remain turned on, once it has been decided that it
            should turn on. Defaults to '1', i.e., not limited to a consecutive number of time steps.
        """
        self.var_name = var_name
        self.lower_activation_limit = lower_activation_limit
        self.var_nom = var_nom
        self.min_off_time = min_off_time
        self.min_on_time = min_on_time

        o.new_var(var_name+"_state", dtype=np.bool_, func=lambda model:
                  abs(o.schedule[self.var_name][o.op_slice]) > abs(0.01 * var_nom))

    def apply(self, m, mode=""):
        if mode == "integer" and self.lower_activation_limit != 0.0 and self.var_nom != 0.0:
            # Add additional binary variables representing operating state
            if hasattr(m, self.var_name + "_state"):
                raise ValueError("model already has a component named: {}".format(self.var_name + "_state"))

            var = pyomo.Var(m.t, domain=pyomo.Binary, initialize=0)
            m.add_component(self.var_name + "_state_vars", var)

            # Couple state to operating variable
            if self.lower_activation_limit == 1.0:
                def p_activation_rule(model, t):
                    orig_var = getattr(m, self.var_name + "_vars")
                    var = getattr(m, self.var_name + "_state_vars")
                    return orig_var[t] == var[t] * self.var_nom

                if hasattr(m, self.var_name + "_activation_constr"):
                    raise ValueError("model already has a component named: {}".
                                     format(self.var_name + "_activation_constr"))
                m.add_component(self.var_name + "_activation_constr", pyomo.Constraint(m.t,
                                                                                       rule=p_activation_rule))
            else:
                def p_state_rule(model, t):
                    orig_var = getattr(m, self.var_name + "_vars")
                    var = getattr(m, self.var_name + "_state_vars")
                    if self.var_nom > 0:
                        return orig_var[t] <= var[t] * self.var_nom
                    else:
                        return orig_var[t] >= var[t] * self.var_nom

                if hasattr(m, self.var_name + "_state_constr"):
                    raise ValueError("model already has a component named: {}".format(self.var_name + "_state_constr"))
                m.add_component(self.var_name + "_state_constr", pyomo.Constraint(m.t, rule=p_state_rule))

                def p_activation_rule(model, t):
                    orig_var = getattr(m, self.var_name + "_vars")
                    var = getattr(m, self.var_name + "_state_vars")
                    if self.var_nom > 0:
                        return orig_var[t] >= var[t] * self.var_nom * self.lower_activation_limit
                    else:
                        return orig_var[t] <= var[t] * self.var_nom * self.lower_activation_limit

                if hasattr(m, self.var_name + "_activation_constr"):
                    raise ValueError("model already has a component named: {}".
                                     format(self.var_name + "_activation_constr"))
                m.add_component(self.var_name + "_activation_constr", pyomo.Constraint(m.t,
                                                                                       rule=p_activation_rule))

            # minimum off-time constraint
            if self.min_off_time > 1:
                m.t_min_off = pyomo.RangeSet(0, m.t.last()-self.min_off_time)

                indicator_1_to_0 = pyomo.Var(m.t_min_off, domain=pyomo.Reals,
                                             bounds=(-constants.INFINITY, constants.INFINITY), initialize=0)
                m.add_component(self.var_name + "_indicator_1_to_0_vars", indicator_1_to_0)

                def indicator_1_to_0_lb_rule(model, t):
                    count_var_1_to_0 = getattr(m, self.var_name + "_indicator_1_to_0_vars")
                    return count_var_1_to_0[t] >= -constants.INFINITY

                if hasattr(m, self.var_name + "_indicator_1_to_0_lb_constr"):
                    raise ValueError("model already has a component named: {}".
                                     format(self.var_name + "_indicator_1_to_0_lb_constr"))
                m.add_component(self.var_name + "_indicator_1_to_0_lb_constr",
                                pyomo.Constraint(m.t_min_off, rule=indicator_1_to_0_lb_rule))

                def indicator_1_to_0_ub_rule(model, t):
                    count_var_1_to_0 = getattr(m, self.var_name + "_indicator_1_to_0_vars")
                    return count_var_1_to_0[t] <= constants.INFINITY

                if hasattr(m, self.var_name + "_indicator_1_to_0_ub_constr"):
                    raise ValueError("model already has a component named: {}".
                                     format(self.var_name + "_indicator_1_to_0_ub_constr"))
                m.add_component(self.var_name + "_indicator_1_to_0_ub_constr",
                                pyomo.Constraint(m.t_min_off, rule=indicator_1_to_0_ub_rule))

                def indicator_1_to_0_rule(model, t):
                    var = getattr(m, self.var_name + "_state_vars")
                    count_var_1_to_0 = getattr(m, self.var_name + "_indicator_1_to_0_vars")
                    if t == 0:
                        return count_var_1_to_0[t] == 0
                    else:
                        return count_var_1_to_0[t] == var[t-1] - var[t]

                if hasattr(m, self.var_name + "_indicator_1_to_0_constr"):
                    raise ValueError("model already has a component named: {}".
                                     format(self.var_name + "_indicator_1_to_0_constr"))
                m.add_component(self.var_name + "_indicator_1_to_0_constr",
                                pyomo.Constraint(m.t_min_off, rule=indicator_1_to_0_rule))

                def min_off_time_rule(model, t):
                    var = getattr(m, self.var_name + "_state_vars")
                    count_var_1_to_0 = getattr(m, self.var_name + "_indicator_1_to_0_vars")
                    return (sum(var[t+tau] for tau in range(self.min_off_time)) <= self.min_off_time *
                            (1.0-count_var_1_to_0[t]))

                if hasattr(m, self.var_name + "_min_off_time_constr"):
                    raise ValueError("model already has a component named: {}".
                                     format(self.var_name + "_min_off_time_constr"))
                m.add_component(self.var_name + "_min_off_time_constr", pyomo.Constraint(m.t_min_off,
                                                                                         rule=min_off_time_rule))

            # minimum on-time constraint
            if self.min_on_time > 1:
                m.t_min_on = pyomo.RangeSet(0, m.t.last() - self.min_on_time)

                indicator_0_to_1 = pyomo.Var(m.t_min_on, domain=pyomo.Reals,
                                             bounds=(-constants.INFINITY, constants.INFINITY), initialize=0)
                m.add_component(self.var_name + "_indicator_0_to_1_vars", indicator_0_to_1)

                def indicator_0_to_1_lb_rule(model, t):
                    count_var_0_to_1 = getattr(m, self.var_name + "_indicator_0_to_1_vars")
                    return count_var_0_to_1[t] >= -constants.INFINITY

                if hasattr(m, self.var_name + "_indicator_0_to_1_lb_constr"):
                    raise ValueError("model already has a component named: {}".
                                     format(self.var_name + "_indicator_0_to_1_lb_constr"))
                m.add_component(self.var_name + "_indicator_0_to_1_lb_constr",
                                pyomo.Constraint(m.t_min_on, rule=indicator_0_to_1_lb_rule))

                def indicator_0_to_1_ub_rule(model, t):
                    count_var_0_to_1 = getattr(m, self.var_name + "_indicator_0_to_1_vars")
                    return count_var_0_to_1[t] <= constants.INFINITY

                if hasattr(m, self.var_name + "_indicator_0_to_1_ub_constr"):
                    raise ValueError("model already has a component named: {}".
                                     format(self.var_name + "_indicator_0_to_1_ub_constr"))
                m.add_component(self.var_name + "_indicator_0_to_1_ub_constr",
                                pyomo.Constraint(m.t_min_on, rule=indicator_0_to_1_ub_rule))

                def indicator_0_to_1_rule(model, t):
                    var = getattr(m, self.var_name + "_state_vars")
                    count_var_0_to_1 = getattr(m, self.var_name + "_indicator_0_to_1_vars")
                    if t == 0:
                        return count_var_0_to_1[t] == 0
                    else:
                        return count_var_0_to_1[t] == var[t] - var[t-1]

                if hasattr(m, self.var_name + "_indicator_0_to_1_constr"):
                    raise ValueError("model already has a component named: {}".
                                     format(self.var_name + "_indicator_0_to_1_constr"))
                m.add_component(self.var_name + "_indicator_0_to_1_constr",
                                pyomo.Constraint(m.t_min_on, rule=indicator_0_to_1_rule))

                def min_on_time_rule(model, t):
                    var = getattr(m, self.var_name + "_state_vars")
                    count_var_0_to_1 = getattr(m, self.var_name + "_indicator_0_to_1_vars")
                    return (sum(var[t+tau] for tau in range(self.min_on_time)) >= self.min_on_time *
                            count_var_0_to_1[t])

                if hasattr(m, self.var_name + "_min_on_time_constr"):
                    raise ValueError("model already has a component named: {}".
                                     format(self.var_name + "_min_on_time_constr"))
                m.add_component(self.var_name + "_min_on_time_constr", pyomo.Constraint(m.t_min_off,
                                                                                        rule=min_on_time_rule))



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import pyomo.environ as pyomo
import pycity_base.classes.supply.thermal_energy_storage as tes

from pycity_scheduling.classes.thermal_entity_cooling import ThermalEntityCooling


class ThermalCoolingStorage(ThermalEntityCooling, tes.ThermalEnergyStorage):
    """
    Extension of pyCity_base class ThermalEnergyStorage for scheduling purposes.
    A thermal cooling storage device can be used as a 'buffer' within a cooling system setup.

    Parameters
    ----------
    environment : Environment
        Common Environment instance.
    e_th_max : float
        Amount of energy the TCS is able to store in [kWh].
    soc_init : float
        Initial state of charge.
    loss_factor : float, optional
        Storage's loss factor (area*U_value) in [W/K].
    storage_end_equality : bool, optional
        Defaults to `False`.
        `True` if the soc at the end of the scheduling has to be equal to
        the initial soc.
        `False` if it has to be greater or equal than the initial soc.


    Notes
    -----
    - TCSs offer sets of constraints for operation. The following constraints
      and bounds are generated by the TCS:

    .. math::
        e_{th\\_cool} &=& e_{th\\_cool\\_previous} * (1-th\\_loss) + p_{th\\_cool} * \\Delta t \\\\
        \\text{with} \\quad e_{th\\_cool\\_previous}
        &=& \\begin{bmatrix} e_{th\\_ini} & e_{th\\_cool\\_0} & \\cdots & e_{th\\_cool\\_n-1}\\end{bmatrix}

    - Additional constraints generated by the parameters are:

    .. math::
        e_{th\\_cool\\_t\\_last} &=& soc\\_init * e_{th\\_max}, & \\quad \\text{if storage_end_equality} \\\\
        e_{th\\_cool\\_t\\_last} &\\geq& soc\\_init * e_{th\\_max}, & \\quad \\text{else}
    """

    def __init__(self, environment, e_th_max, soc_init=0.5, loss_factor=0,
                 storage_end_equality=False):
        # Room temperature of 20 C and flow temperature of 55 C
        capacity = e_th_max / self.c_water / 15 * 3.6e6
        super().__init__(environment, 5, capacity, 5, 20, loss_factor)
        self._long_id = "TCS_" + self._id_string

        self.e_th_max = e_th_max
        self.soc_init = soc_init
        self.storage_end_equality = storage_end_equality

        self.th_loss_coeff = (self.k_losses / self.capacity / self.c_water * self.time_slot * 3600)

        self.new_var("e_th_cool")

    def populate_model(self, model, mode="convex"):
        """
        Add device block to pyomo ConcreteModel

        Call parent's `populate_model` method and set variables lower bounds
        to `None`. Then add variables for the state of charge with an upper
        bound of `self.e_th_max`. Also add continuity constraints to the model.

        Parameters
        ----------
        model : pyomo.ConcreteModel
        mode : str, optional
            Specifies which set of constraints to use.

            - `convex`  : Use linear constraints
            - `integer` : Use same constraints as convex mode
        """
        super().populate_model(model, mode)
        m = self.model

        if mode == "convex" or "integer":
            m.e_th_cool_vars = pyomo.Var(m.t, domain=pyomo.Reals, initialize=0)
            m.e_th_init = pyomo.Param(default=self.soc_init * self.e_th_max, mutable=True)

            def e_th_cool_vars_lb_rule(model, t):
                return model.e_th_cool_vars[t] >= 0.0
            m.e_th_cool_vars_lb_constr = pyomo.Constraint(m.t, rule=e_th_cool_vars_lb_rule)

            def e_th_cool_vars_ub_rule(model, t):
                return model.e_th_cool_vars[t] <= self.e_th_max
            m.e_th_cool_vars_ub_constr = pyomo.Constraint(m.t, rule=e_th_cool_vars_ub_rule)

            def e_rule(model, t):
                e_th_last = model.e_th_cool_vars[t-1] if t >= 1 else model.e_th_init
                return (model.e_th_cool_vars[t] == e_th_last * (1 - self.th_loss_coeff) +
                        m.p_th_cool_vars[t] * self.time_slot)
            m.e_constr = pyomo.Constraint(m.t, rule=e_rule)

            def e_end_rule(model):
                if self.storage_end_equality:
                    return model.e_th_cool_vars[self.op_horizon-1] == self.e_th_max * self.soc_init
                else:
                    return model.e_th_cool_vars[self.op_horizon-1] >= self.e_th_max * self.soc_init
            m.e_end_constr = pyomo.Constraint(rule=e_end_rule)

        else:
            raise ValueError(
                "Mode %s is not implemented by class ThermalCoolingStorage." % str(mode)
            )

    def update_model(self, mode=""):
        m = self.model
        timestep = self.timestep

        if timestep == 0:
            m.e_th_init = self.soc_init * self.e_th_max
        else:
            m.e_th_init = self.e_th_cool_schedule[timestep-1]
        return



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import pyomo.environ as pyomo
import matplotlib.pyplot as plt

import ray
import copy
import psutil

from pycity_scheduling.classes import (CityDistrict, Building, Photovoltaic, WindEnergyConverter)
from pycity_scheduling.algorithms.algorithm import IterationAlgorithm, DistributedAlgorithm
from pycity_scheduling.algorithms.algorithm_ray import RayADMMMIQPSolverNode, list_into_n_chunks
from pycity_scheduling.solvers import DEFAULT_SOLVER, DEFAULT_SOLVER_OPTIONS


class ExchangeMIQPADMMRay(IterationAlgorithm, DistributedAlgorithm):
    """Implementation of the Exchange MIQP ADMM algorithm.

    Implements the Exchange MIQP ADMM algorithm described in [1], [2], and [3] using parallel computations with Ray.

    Parameters
    ----------
    city_district : CityDistrict
    solver : str, optional
        Solver to use for solving (sub)problems.
    solver_options : dict, optional
        Options to pass to calls to the solver. Keys are the name of
        the functions being called and are one of `__call__`, `set_instance_`,
        `solve`.
        `__call__` is the function being called when generating an instance
        with the pyomo SolverFactory. In addition to the options provided,
        `node_ids` is passed to this call containing the IDs of the entities
        being optimized.
        `set_instance` is called when a pyomo Model is set as an instance of
        a persistent solver. `solve` is called to perform an optimization. If
        not set, `save_results` and `load_solutions` may be set to false to
        provide a speedup.
    mode : str, optional
        Specifies which set of constraints to use.
        - `convex`  : Use linear constraints
        - `integer`  : May use non-linear constraints
    eps_exch_primal : float, optional
        Primal stopping criterion for the ADMM algorithm.
    eps_exch_dual : float, optional
        Dual stopping criterion for the ADMM algorithm.
    gamma : float, optional
        Exchange MIQP ADMM scaling parameter
    gamma_incr : float, optional
        Varying scaling parameter scheme increase parameter
    rho : float, optional
        Stepsize for the ADMM algorithm.
    varying_penalty_parameter : bool, optional
        Apply a varying penalty parameter scheme, see [3] - chapter 3.4.1
    tau_incr : float, optional
        Varying penalty parameter scheme increase parameter
    tau_decr : float, optional
        Varying penalty parameter scheme decrease parameter
    mu : float, optional
        Varying penalty parameter scheme conditional change parameter
    max_iterations : int, optional
        Maximum number of ADMM iterations.
    robustness : tuple, optional
        Tuple of two floats. First entry defines how many time steps are
        protected from deviations. Second entry defines the magnitude of
        deviations which are considered.
    ray_cpu_count : int, optional
        Number of CPU cores to be used by ray for parallelization.
        Default: Detect the number of CPUs automatically.

    References
    ----------
    [1] "Alternating Direction Method of Multipliers for Decentralized
    Electric Vehicle Charging Control" by Jose Rivera, Philipp Wolfrum,
    Sandra Hirche, Christoph Goebel, and Hans-Arno Jacobsen
    Online: https://mediatum.ub.tum.de/doc/1187583/1187583.pdf (accessed on 2020/09/28)

    [2] "A simple effective heuristic for embedded mixed-integer quadratic programming" by Reza Takapoui,
    Nicholas Moehle, Stephen Boyd, and Alberto Bemporad
    Online: https://web.stanford.edu/~boyd/papers/pdf/miqp_admm.pdf (accessed on 2023/09/06)

    [3] "Distributed Optimization and Statistical Learning via the Alternating Direction Method of Multipliers" by
    Stephen Boyd, Neal Parikh, Eric Chu, Borja Peleato, and Jonathan Eckstein
    Online: https://web.stanford.edu/~boyd/papers/pdf/admm_distr_stats.pdf (accessed on 2023/10/09)
    """
    def __init__(self, city_district, solver=DEFAULT_SOLVER, solver_options=DEFAULT_SOLVER_OPTIONS,
                 mode="convex", x_update_mode="constrained", eps_exch_primal=0.01, eps_exch_dual=0.1, gamma=1.0,
                 gamma_incr=1.0, rho=2.0, varying_penalty_parameter=False, tau_incr=2.0, tau_decr=2.0, mu=10.0,
                 max_iterations=10000, robustness=None, ray_cpu_count=None):
        super(ExchangeMIQPADMMRay, self).__init__(city_district, solver, solver_options, mode)

        self.x_update_mode = x_update_mode
        self.eps_exch_primal = eps_exch_primal
        self.eps_exch_dual = eps_exch_dual
        self.gamma = gamma
        self.gamma_incr = gamma_incr
        self.rho = rho
        self.varying_penalty_parameter = varying_penalty_parameter
        self.tau_incr = tau_incr
        self.tau_decr = tau_decr
        self.mu = mu
        self.max_iterations = max_iterations
        self.op_horizon = self.city_district.op_horizon

        # Only consider entities of type CityDistrict, Building, Photovoltaic, WindEnergyConverter
        self._entities = [entity for entity in self.entities if
                          isinstance(entity, (CityDistrict, Building, Photovoltaic, WindEnergyConverter))]

        # Initialize ray before using ray functions
        self.ray_cpu_count = ray_cpu_count

        # Ray initialization on a single machine vs. on an active ray cluster:
        try:
            ray.shutdown()
            if self.ray_cpu_count is not None:
                ray.init(ignore_reinit_error=True, num_cpus=int(self.ray_cpu_count))
            else:
                ray.init(ignore_reinit_error=True, num_cpus=min(len(self._entities), psutil.cpu_count(logical=False)))
        except:
            ray.init(ignore_reinit_error=True)
        resources = ray.cluster_resources()

        # Create a remote ray actor solver node for each available CPU core
        node_num = len(ray.nodes())
        if self.ray_cpu_count is not None:
            cpu_count = self.ray_cpu_count
        else:
            cpu_count = int(max(resources["CPU"], psutil.cpu_count(logical=False)))
        entity_chunks = copy.deepcopy(list(list_into_n_chunks(self._entities,
                                                              int(min(cpu_count, len(self._entities))))))
        print("{} nodes available | {} CPUs available/defined | {} entities are present | {} entity chunks created".
            format(node_num, cpu_count, len(self._entities), len(entity_chunks)))

        self.nodes = []
        self.global_entity_indices = dict()

        for i, chunk in zip(range(len(entity_chunks)), entity_chunks):
            for chunk_entity in chunk:
                # Find index of entity in self._entities and safe it to the dictionary
                global_entity_index = (
                    next((i for i, item in enumerate(self._entities) if item.id == chunk_entity.id), -1))
                self.global_entity_indices.update({chunk_entity.id: global_entity_index})
            self.nodes.append(RayADMMMIQPSolverNode.remote(solver, 
                                                           solver_options, 
                                                           entities=chunk,
                                                           gamma=self.gamma,
                                                           rho=self.rho,
                                                           node_index=i,
                                                           max_iterations=self.max_iterations, 
                                                           op_horizon=self.op_horizon, 
                                                           mode=mode,
                                                           x_update_mode=self.x_update_mode, 
                                                           robustness=robustness))

        # Additions to the Exchange ADMM algorithm in order to obtain Exchange MIQP ADMM
        self.feasible = False

        # Set parameters on all remotes
        set_param_refs = [node.set_model_parameters.remote() for node in self.nodes]

        # Wait until set parameters is finished on all nodes
        while len(set_param_refs) > 0:
            done_set_param_refs, set_param_refs = ray.wait(set_param_refs)
            for ref in done_set_param_refs:
                ray.get(ref)
        
        self._add_objective(exchange_admm_obj_terms=True, miqp_admm_obj_terms=True)

    # Returns True, if no constraint or binary variable violations occur
    def _check_violations(self, print_out=True, detailed_print_out=False):
        constr_counter = 0
        var_counter = 0
        number_constraints = 0
        number_binaries = 0
        # Get parameters from all remote nodes
        get_counter_refs = [node.check_violations.remote(detailed_print_out=detailed_print_out) for node in self.nodes]
        # Wait until get parameters is finished on all nodes
        while len(get_counter_refs):
            done_get_counter_refs, get_counter_refs = ray.wait(get_counter_refs)
            for done_get_counter_ref in done_get_counter_refs:
                returns = ray.get(done_get_counter_ref)
                constr_counter += returns[0]
                var_counter  += returns[1]
                number_constraints += returns[2]
                number_binaries  += returns[3]

        if print_out:
            print("Violated constraints:", constr_counter,  "of in total", number_constraints)
            print("Violated binaries:", var_counter, "of in total", number_binaries)
        if var_counter == 0 and constr_counter == 0:
            return True
        else:
            return False

    # Returns the objective value
    def _get_objective(self):
        # Get parameters from all remote nodes
        get_obj_refs = [node.get_objective.remote() for node in self.nodes]
        obj_value = 0
        # Wait until get parameters is finished on all nodes
        while len(get_obj_refs) > 0:
            done_get_obj_refs, get_obj_refs = ray.wait(get_obj_refs)
            for ref in done_get_obj_refs:
                remote_obj_value = ray.get(ref)
                obj_value += remote_obj_value
        return obj_value

    def _add_objective(self, exchange_admm_obj_terms=True, miqp_admm_obj_terms=True):
        # Add objective on all remotes
        add_obj_refs = [node.add_objective.remote(exchange_admm_obj_terms, miqp_admm_obj_terms) for node in self.nodes]
        # Wait until objectives have been set on all nodes
        while len(add_obj_refs) > 0:
            done_add_obj_refs, add_obj_refs = ray.wait(add_obj_refs)
            for ref in done_add_obj_refs:
                returned = ray.get(ref)
        return

    def _presolve(self, full_update, beta, robustness, debug):
        results, params = super()._presolve(full_update, beta, robustness, debug)

        set_model_beta_refs = []
        # Set the model betas on each node
        for i, node in enumerate(self.nodes):
            set_model_beta_refs.append(node.set_model_betas.remote(params))
            if full_update:
                set_model_beta_refs.append(node.full_update.remote(robustness))

        while len(set_model_beta_refs):
            done_set_model_beta_refs, set_model_beta_refs = ray.wait(set_model_beta_refs)
            for ref in done_set_model_beta_refs:
                ray.get(ref)

        results["r_norms"] = []
        results["s_norms"] = []
        results["gamma_value"] = []
        results["rho_value"] = []
        return results, params

    def _postsolve(self, results, params, debug):
        # ------------------------------------------
        # 1) Remote post solve
        # ------------------------------------------
        # Execute a remote version of the postsolve method on each node,
        # the local schedules of each entity are the calculated
        ray.get([node.postsolve.remote() for node in self.nodes])

        schedule_dict = dict()
        get_schedules_refs = [node.get_all_schedules.remote() for node in self.nodes]
        # ------------------------------------------
        # 2) Wait for remote post solves
        # ------------------------------------------
        # Wait until all nodes are finished and retrive there local dictionary
        # The dictionary contains each asset id with its corresponding schedules
        while len(get_schedules_refs) > 0:
            done_get_schedules_refs, get_schedules_refs = ray.wait(get_schedules_refs) 
            for done_schedule in done_get_schedules_refs:
                node_schedules = ray.get(done_schedule)
                schedule_dict.update(node_schedules)

        # ------------------------------------------
        # 3) Local schedule update
        # ------------------------------------------
        # Update the local entities with the schedules calculated on the remote nodes
        for entity in self._entities:
            if not isinstance(entity, CityDistrict):
                for asset in entity.get_all_entities():
                    if hasattr(asset, "id"):
                        schedules = schedule_dict.get(asset.id)
                        if schedules is not None:
                            asset.schedules = schedules
            if hasattr(entity, "id"):
                schedules = schedule_dict.get(entity.id)
                if schedules is not None:
                    entity.schedules = schedules

        city_district = self._entities[0]
        for entity in city_district.get_all_entities():
            if hasattr(entity, "id"):
                schedules = schedule_dict.get(entity.id)
                if schedules is not None:
                    entity.schedules = schedules

        # Terminate all ray actors
        for node in self.nodes:
            node.exit.remote()
        return

    def _is_last_iteration(self, results, params, debug):
        if super(ExchangeMIQPADMMRay, self)._is_last_iteration(results, params, debug):
            self._check_violations()
            return True
        if results["r_norms"][-1] <= self.eps_exch_primal and results["s_norms"][-1] <= self.eps_exch_dual:
            if self.feasible:
                self._check_violations()
                return True
            else:
                return False

    def _iteration(self, results, params, debug):
        super(ExchangeMIQPADMMRay, self)._iteration(results, params, debug)
        op_horizon = self._entities[0].op_horizon

        # fill parameters if not already present
        if "p_el" not in params:
            params["p_el"] = np.zeros((len(self._entities), op_horizon))
        if "x_" not in params:
            params["x_"] = np.zeros(op_horizon)
        if "u" not in params:
            params["u"] = np.zeros(op_horizon)
        last_u = params["u"]
        last_p_el = params["p_el"]
        last_x_ = params["x_"]

        # ------------------------------------------
        # 0) Storage for variables
        # ------------------------------------------
        p_el_schedules_list = [None] * len(self._entities)
        last_p_el_stored = ray.put(last_p_el)
        last_x__stored = ray.put(last_x_)
        last_u_stored = ray.put(last_u)
        global_entity_indices_stored = ray.put(self.global_entity_indices)

        # ------------------------------------------
        # 1) Optimize all entities
        # ------------------------------------------
        solve_ids = [node.solve.remote(variable_refs=[last_p_el_stored, last_x__stored, last_u_stored,
                                                      global_entity_indices_stored], update_constr=True, debug=debug)
                     for i, node in zip(range(len(self.nodes)), self.nodes)]

        obj_value = 0
        while len(solve_ids):
            done_solve_ids, solve_ids = ray.wait(solve_ids)
            for done_solve_id in done_solve_ids:
                returns = ray.get(done_solve_id) 
                obj_value += returns[2]
                for entity_id, entity_value in zip(returns[0], returns[1]):
                    global_index = self.global_entity_indices[entity_id]
                    p_el_schedules_list[global_index] = entity_value

        # ------------------------------------------
        # 2) Calculate incentive signal update
        # ------------------------------------------
        p_el_schedules = np.array(p_el_schedules_list)
        x_ = (-p_el_schedules[0] + sum(p_el_schedules[1:])) / len(self._entities)
        results["r_norms"].append(np.sqrt(len(self._entities)) * np.linalg.norm(x_))

        s = np.zeros_like(p_el_schedules)
        s[0] = - self.rho * (-p_el_schedules[0] + last_p_el[0] + last_x_ - x_)
        for i in range(1, len(self._entities)):
            s[i] = - self.rho * (p_el_schedules[i] - last_p_el[i] + last_x_ - x_)
        results["s_norms"].append(np.linalg.norm(s.flatten()))

        results["obj_value"].append(self._get_objective())

        # To check the latest solution for feasibility, fix all binary vars at either 0 or 1 and try to recalculate the
        # schedule based on an additional optimization. Of course, there are other and possibly more
        # straightforward ways to check for feasibility. This will be future work.
        if ((results["r_norms"][-1] <= self.eps_exch_primal) and (results["s_norms"][-1] <= self.eps_exch_dual)) or \
                (results["iterations"][-1] >= self.max_iterations):
            if results["iterations"][-1] >= self.max_iterations:
                print("Iteration " + str(results["iterations"][-1]) + ": " +
                      "Maximum iteration limit reached. Checking the current solution for feasibility.")
            else:
                print("Iteration " + str(results["iterations"][-1]) + ": " +
                      "Stopping criteria satisfied. Checking the current solution for feasibility.")
            self._add_objective(exchange_admm_obj_terms=True, miqp_admm_obj_terms=True)
            self.fix_variables()
            if self.x_update_mode == "unconstrained":
                self.activate_constraints()

            solve_ids = [node.solve.remote(variable_refs=[last_p_el_stored, last_x__stored, last_u_stored,
                                                          global_entity_indices_stored], update_constr=True,
                                           debug=False) for i, node in zip(range(len(self.nodes)), self.nodes)]
            self.feasible = True
            while len(solve_ids):
                done_solve_ids, solve_ids = ray.wait(solve_ids)
                for done_solve_id in done_solve_ids:
                    returns = ray.get(done_solve_id) 
                    self.feasible = self.feasible and returns[3]

            if self.feasible:
                print("Success. The solution is feasible!")
                self.release_variables()
                if self.x_update_mode == "unconstrained":
                    self.deactivate_constraints()
            else:
                print("Failure. The solution is infeasible, because at least one subsystem is (still) infeasible!")
                if results["iterations"][-1] >= self.max_iterations:
                    print("Attention: The final solution could be infeasible!")
                else:
                    self._add_objective(exchange_admm_obj_terms=True, miqp_admm_obj_terms=True)
                    self.release_variables()
                    if self.x_update_mode == "unconstrained":
                        self.deactivate_constraints()
                    solve_ids = [node.solve.remote(variable_refs=[last_p_el_stored, last_x__stored, last_u_stored,
                                                                  global_entity_indices_stored], update_constr=True,
                                                   debug=False) for i, node in zip(range(len(self.nodes)), self.nodes)]
                    while len(solve_ids):
                        done_solve_ids, solve_ids = ray.wait(solve_ids)
                        for done_solve_id in done_solve_ids:
                            returns = ray.get(done_solve_id)
                self.gamma = min(float(self.gamma_incr)*self.gamma, 10e9)
                ray.get([node.update_gamma.remote(self.gamma) for node in self.nodes])
        results["gamma_value"].append(self.gamma)

        # ----------------------------------------
        # 3) incentive signal update (MIQP ADMM)
        # ----------------------------------------
        incentive_refs = [node.incentive_signal_update.remote(results["iterations"][-1]-1) for node in self.nodes]

        while len(incentive_refs) > 0:
            done_incentive_refs, incentive_refs = ray.wait(incentive_refs)
            for ref in done_incentive_refs:
                result = ray.get(ref)

        # if desired, vary the penalty parameter
        if self.varying_penalty_parameter:
            if results["r_norms"][-1] > self.mu * results["s_norms"][-1]:
                self.rho *= float(self.tau_incr)
            elif results["s_norms"][-1] > self.mu * results["r_norms"][-1]:
                self.rho /= float(self.tau_decr)
            ray.get([node.update_rho.remote(self.rho) for node in self.nodes])
        results["rho_value"].append(self.rho)

        # save parameters for another iteration
        params["p_el"] = p_el_schedules
        params["x_"] = x_
        params["u"] += x_
        return

    def fix_variables(self):
        fix_vars_refs = [node.fix_variables.remote() for node in self.nodes]
        while len(fix_vars_refs) > 0:
            done_fix_vars_refs, fix_vars_refs = ray.wait(fix_vars_refs)
            for ref in done_fix_vars_refs:
                ray.get(ref)
        return

    def release_variables(self):
        release_vars_refs = [node.release_variables.remote() for node in self.nodes]
        while len(release_vars_refs) > 0:
            done_release_vars_refs, release_vars_refs = ray.wait(release_vars_refs)
            for ref in done_release_vars_refs:
                ray.get(ref)
        return

    def activate_constraints(self):
        # Set parameters on all remotes
        activate_contraints_refs = [node.activate_constraints.remote() for node in self.nodes]

        # Wait until set parameters is finished on all nodes
        while len(activate_contraints_refs) > 0:
            done_activate_contraints_refs, activate_contraints_refs = ray.wait(activate_contraints_refs)
            for ref in done_activate_contraints_refs:
                ray.get(ref)
        return

    def deactivate_constraints(self):
        # Set parameters on all remotes
        deactivate_contraints_refs = [node.deactivate_constraints.remote() for node in self.nodes]

        # Wait until set parameters is finished on all nodes
        while len(deactivate_contraints_refs) > 0:
            done_deactivate_contraints_refs, deactivate_contraints_refs = ray.wait(deactivate_contraints_refs)
            for ref in done_deactivate_contraints_refs:
                ray.get(ref)
        return



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import pyomo.environ as pyomo

from pycity_scheduling.classes import *
from pycity_scheduling.algorithms import *
from pycity_scheduling.solvers import *


# This examples demonstrates the warmstart capability for algorithms implemented in pycity_scheduling.


def main(do_plot=False):
    print("\n\n------ Example 08: Algorithm Warmstart ------\n\n")

    # Unfortunately, algorithm warmstart capabilities are supported by the Gurobi/CPLEX solvers only:
    if not (DEFAULT_SOLVER == "gurobi_direct" or DEFAULT_SOLVER == "gurobi_persistent"):
        print("Algorithm warmstart is currently supported by the Gurobi solver only! Example is not executed.")
        return

    # Define timer, price, weather, and environment objects:
    t = Timer()
    p = Prices(timer=t)
    w = Weather(timer=t)
    e = Environment(timer=t, weather=w, prices=p)

    # City district with district operator objective "peak-shaving":
    cd = CityDistrict(environment=e, objective='peak-shaving')

    # Schedule two sample buildings. The buildings' objectives are defined as "peak-shaving".

    # Building no. one comes with fixed load, space heating, electric heater, pv unit, thermal energy storage, and
    # electrical battery energy storage:
    bd1 = Building(environment=e, objective='peak-shaving')
    cd.addEntity(entity=bd1, position=[0, 0])
    bes = BuildingEnergySystem(environment=e)
    bd1.addEntity(bes)
    ths = ThermalHeatingStorage(environment=e, e_th_max=40, soc_init=0.5)
    bes.addDevice(ths)
    eh = ElectricHeater(environment=e, p_th_nom=10, lower_activation_limit=0.25)
    bes.addDevice(eh)
    ap = Apartment(environment=e)
    bd1.addEntity(ap)
    load = np.array(-np.sin(np.linspace(0, 2.1*np.pi, 96)+1))
    fi = FixedLoad(e, method=0, demand=load)
    ap.addEntity(fi)
    sh = SpaceHeating(environment=e, method=0, loadcurve=load)
    ap.addEntity(sh)
    pv = Photovoltaic(environment=e, method=1, peak_power=4.6)
    bes.addDevice(pv)
    bat = Battery(environment=e, e_el_max=4.8, p_el_max_charge=3.6, p_el_max_discharge=3.6, eta=1.0)
    bes.addDevice(bat)

    # Building no. two comes with deferrable load, curtailable load, space heating, chp unit, thermal energy storage,
    # and an electric vehicle:
    bd2 = Building(environment=e, objective='peak-shaving')
    cd.addEntity(entity=bd2, position=[0, 1])
    bes = BuildingEnergySystem(environment=e)
    bd2.addEntity(bes)
    ths = ThermalHeatingStorage(environment=e, e_th_max=35, soc_init=0.5)
    bes.addDevice(ths)
    chp = CombinedHeatPower(environment=e, p_th_nom=20.0, lower_activation_limit=0.0)
    bes.addDevice(chp)
    ap = Apartment(environment=e)
    bd2.addEntity(ap)
    load = np.array(-2*np.sin(np.linspace(0, 1.9*np.pi, 96)+3))
    dl = DeferrableLoad(environment=e, p_el_nom=2.0, e_consumption=2.0)
    ap.addEntity(dl)
    cl = CurtailableLoad(environment=e, p_el_nom=1.6, max_curtailment=0.8)
    ap.addEntity(cl)
    sh = SpaceHeating(environment=e, method=0, loadcurve=load)
    ap.addEntity(sh)
    ev = ElectricVehicle(environment=e, e_el_max=37.0, p_el_max_charge=22.0, soc_init=0.5, simulate_driving=False,
                         minimum_soc_end=0.8, charging_time=np.ones(t.simu_horizon))
    ap.addEntity(ev)

    # Perform the scheduling with the Exchange ADMM algorithm to obtain an algorithm warmstart point:
    opt = ExchangeADMM(city_district=cd, rho=2.0, eps_primal=0.01, eps_dual=0.01, mode="integer")
    r1 = opt.solve()
    imbalance = sum(np.abs(cd.schedule["p_el"] - bd1.schedule["p_el"] - bd2.schedule["p_el"]))

    # Let the city district account for power imbalances in order to achieve a feasible schedule.
    cd.account_imbalance()
    admm_obj_imbalance = pyomo.value(cd.get_objective() + bd1.get_objective() + bd2.get_objective())

    # Load the obtained values into the model to influence the pyomo.value parameter:
    cd.load_schedule_into_model()
    admm_obj = pyomo.value(cd.get_objective() + bd1.get_objective() + bd2.get_objective())

    # Now perform a central optimization with the given warmstart point:
    opt = CentralOptimization(city_district=cd, mode="integer",
                              solver_options={'solve': {'warmstart': True, 'tee': True}})

    # Now, explicitly load the current schedule into the model:
    cd.load_schedule_into_model()
    r2 = opt.solve()
    central_obj = pyomo.value(cd.get_objective() + bd1.get_objective() + bd2.get_objective())

    # For benchmarking, now perform the central optimization without the warmstart point, too:
    opt = CentralOptimization(city_district=cd, mode="integer")
    cd.load_schedule_into_model()
    r3 = opt.solve()

    admm_parallel_time = r1["times"][-1] - sum(sum(dt.values()) - max(dt.values()) for dt in r1["distributed_times"])
    print("Iterations of ADMM algorithm:                       {: >8}".format(r1["iterations"][-1]))
    print("Total time of ADMM algorithm:                       {: >8.3f}".format(r1["times"][-1]))
    print("Time of ADMM algorithm in 'parallel':               {: >8.3f}".format(admm_parallel_time))
    print("Remaining imbalance of ADMM algorithm:              {: >8.3f}".format(imbalance))
    print("Time of central algorithm with ADMM warmstart:      {: >8.3f}".format(r2["times"][-1]))
    print("Time of central algorithm without ADMM warmstart:   {: >8.3f}".format(r3["times"][-1]))
    print("Objective value of ADMM algorithm with imbalance:   {: >8.3f}".format(admm_obj_imbalance))
    print("Objective value of ADMM algorithm:                  {: >8.3f}".format(admm_obj))
    print("Objective value of central algorithm:               {: >8.3f}".format(central_obj))
    return


# Conclusions:
# The algorithm warm-start capability is a good opportunity to increase the solution speed especially for centralized
# optimization algorithms, as all variables are already pre-initialized with a solution that is located "close" to the
# actual optimum. In particular, using the warm-start capability is reasonable for large problems (i.e., city districts
# that may contain hundreds of buildings and assets).


if __name__ == '__main__':
    # Run example:
    main(do_plot=True)



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import pyomo.environ as pyomo

from pycity_scheduling.algorithms.algorithm import OptimizationAlgorithm, SolverNode
from pycity_scheduling.solvers import DEFAULT_SOLVER, DEFAULT_SOLVER_OPTIONS


class CentralOptimization(OptimizationAlgorithm):
    """
    Implementation of the reference Central Optimization algorithm.

    Only creates one model and calls the solver once. Combines the objectives of all entities.
    """
    def __init__(self, city_district, solver=DEFAULT_SOLVER, solver_options=DEFAULT_SOLVER_OPTIONS, mode="convex",
                 robustness=None):
        super(CentralOptimization, self).__init__(city_district, solver, solver_options, mode)
        # create single solver node for all entities
        self.node = SolverNode(solver, solver_options, self.entities, mode, robustness=robustness)
        # add coupling between all entities in the single model
        self._add_coupling(self.node.model)
        # create pyomo parameters for beta of each entity
        for entity in self.entities:
            entity.model.beta = pyomo.Param(mutable=True, initialize=1)
        self._add_objective()

    def _add_coupling(self, model):
        """Adds the coupling constraint to the model.

        This constraint 'connects' the district operator to all its nodes.

        Parameters
        ----------
        model : pyomo.ConcreteModel
            Model to add the constraint to.
        """
        def p_el_couple_rule(model, t):
            return self.city_district.model.p_el_vars[t] == \
                   sum(entity.model.p_el_vars[t] for entity in self.entities[1:])
        model.couple = pyomo.Constraint(self.city_district.model.t, rule=p_el_couple_rule)
        return

    def _add_objective(self):
        self.node.model.o = pyomo.Objective(expr=sum(entity.model.beta * entity.get_objective()
                                                     for entity in self.entities))
        return

    def _presolve(self, full_update, beta, robustness, debug):
        results, params = super()._presolve(full_update, beta, robustness, debug)
        for entity in self.entities:
            entity.model.beta = self._get_beta(params, beta)
        if full_update:
            self.node.full_update(robustness=robustness)
        else:
            self.node.obj_update()
        return results, params

    def _solve(self, results, params, debug):
        self.node.solve(debug=debug)
        self._save_time(results, params)
        results["obj_value"] = pyomo.value(self.node.model.o)
        return



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


from pycity_scheduling.classes.timer import Timer
from pycity_scheduling.classes.prices import Prices
from pycity_scheduling.classes.weather import Weather
from pycity_scheduling.classes.environment import Environment
from pycity_scheduling.classes.optimization_entity import OptimizationEntity
from pycity_scheduling.classes.thermal_entity_heating import ThermalEntityHeating
from pycity_scheduling.classes.thermal_entity_cooling import ThermalEntityCooling
from pycity_scheduling.classes.electrical_entity import ElectricalEntity
from pycity_scheduling.classes.city_district import CityDistrict
from pycity_scheduling.classes.building import Building
from pycity_scheduling.classes.building_energy_system import BuildingEnergySystem
from pycity_scheduling.classes.apartment import Apartment
from pycity_scheduling.classes.battery import Battery
from pycity_scheduling.classes.boiler import Boiler
from pycity_scheduling.classes.chiller import Chiller
from pycity_scheduling.classes.combined_heat_power import CombinedHeatPower
from pycity_scheduling.classes.curtailable_load import CurtailableLoad
from pycity_scheduling.classes.deferrable_load import DeferrableLoad
from pycity_scheduling.classes.electric_heater import ElectricHeater
from pycity_scheduling.classes.electric_vehicle import ElectricVehicle
from pycity_scheduling.classes.fixed_load import FixedLoad
from pycity_scheduling.classes.heat_pump import HeatPump
from pycity_scheduling.classes.photovoltaic import Photovoltaic
from pycity_scheduling.classes.space_cooling import SpaceCooling
from pycity_scheduling.classes.space_heating import SpaceHeating
from pycity_scheduling.classes.thermal_heating_storage import ThermalHeatingStorage
from pycity_scheduling.classes.thermal_cooling_storage import ThermalCoolingStorage
from pycity_scheduling.classes.wind_energy_converter import WindEnergyConverter
from pycity_scheduling.classes.electrical_grid import ElectricalGrid
from pycity_scheduling.classes.electrical_line import ElectricalLine
from pycity_scheduling.classes.electrical_node import ElectricalNode



__all__ = [
    'Timer',
    'Prices',
    'Weather',
    'Environment',
    'OptimizationEntity',
    'ThermalEntityCooling',
    'ThermalEntityHeating',
    'ElectricalEntity',
    'CityDistrict',
    'Building',
    'BuildingEnergySystem',
    'Apartment',
    'Battery',
    'Boiler',
    'Chiller',
    'CombinedHeatPower',
    'CurtailableLoad',
    'DeferrableLoad',
    'ElectricHeater',
    'ElectricVehicle',
    'ElectricalGrid',
    'ElectricalLine',
    'ElectricalNode',
    'FixedLoad',
    'HeatPump',
    'Photovoltaic',
    'SpaceCooling',
    'SpaceHeating',
    'ThermalHeatingStorage',
    'ThermalCoolingStorage',
    'WindEnergyConverter',
    'all_entities',
    'heating_devices',
    'consumption_devices',
    'generation_devices',
    'storage_devices',
    'filter_entities',
]


all_entities = {
    'OE': OptimizationEntity,
    'TEC': ThermalEntityCooling,
    'TEH': ThermalEntityHeating,
    'EE': ElectricalEntity,
    'CD': CityDistrict,
    'BD': Building,
    'BES': BuildingEnergySystem,
    'AP': Apartment,
    'BAT': Battery,
    'BL': Boiler,
    'CH': Chiller,
    'CHP': CombinedHeatPower,
    'CL': CurtailableLoad,
    'DL': DeferrableLoad,
    'EH': ElectricHeater,
    'EV': ElectricVehicle,
    'FL': FixedLoad,
    'HP': HeatPump,
    'PV': Photovoltaic,
    'SC': SpaceCooling,
    'SH': SpaceHeating,
    'THS': ThermalHeatingStorage,
    'TCS': ThermalCoolingStorage,
    'WEC': WindEnergyConverter,
}

heating_devices = {
    'HP': HeatPump,
    'BL': Boiler,
    'CHP': CombinedHeatPower,
    'EH': ElectricHeater,
}

cooling_devices = {
    'CH': Chiller,
}

consumption_devices = {
    'CL': CurtailableLoad,
    'DL': DeferrableLoad,
    'EV': ElectricVehicle,
    'FL': FixedLoad,
}

generation_devices = {
    'CHP': CombinedHeatPower,
    'PV': Photovoltaic,
    'WEC': WindEnergyConverter,
}

storage_devices = {
    'BAT': Battery,
    'THS': ThermalHeatingStorage,
    'TCS': ThermalCoolingStorage,
}


def filter_entities(entities, entity_type):
    """
    Filter a list of entities for given entity types.

    Parameters
    ----------
    entities : list or generator or
               pycity_scheduling.classes.OptimizationEntity
        Entities to be filtered. When an Optimization Entity is given its
        `get_entities` method is called.
    entity_type : str or list or dict
        In case of `str`: must be the name of a dictionary from this module
                          (e.g. 'generation_devices') or the short name of an entity
                          (e.g. 'PV')
        In case of `list`: must be a list of classes to be filtered for or a
                           list of short names (e.g. `['PV', 'WEC']`)
        In case of `dict`: its values must be the classes to be filtered for

    Yields
    ------
    OptimizationEntity :
        All entities matching the given filter
    """
    if isinstance(entities, OptimizationEntity):
        entities = entities.get_entities()

    if isinstance(entity_type, str):
        if entity_type in all_entities:
            entity_type = [all_entities[entity_type]]
        elif entity_type in __all__:
            entity_type = eval(entity_type)
        else:
            raise ValueError
    if isinstance(entity_type, list):
        if isinstance(entity_type[0], str):
            cls_list = tuple(all_entities[e] for e in entity_type)
        elif issubclass(entity_type[0], OptimizationEntity):
            cls_list = tuple(entity_type)
        else:
            raise ValueError()
    elif isinstance(entity_type, dict):
        cls_list = tuple(entity_type.values())
    else:
        raise ValueError()

    for e in entities:
        if isinstance(e, cls_list):
            yield e



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import os
import os.path as op
import matplotlib.pyplot as plt

from pycity_scheduling.classes import Battery, ElectricVehicle, CombinedHeatPower


_known_varnames = {
    "p_el": plt.get_cmap("tab20").colors[:2],
    "e_el": plt.get_cmap("tab20").colors[8:10],
    "p_th": plt.get_cmap("tab20").colors[2:4],
    "e_th": plt.get_cmap("tab20").colors[6:8]
}

_other_colors = [plt.get_cmap("tab20").colors[i:2+i] for i in range(10, 20, 2)]


def _extended_schedule(var_name, schedule):
    if var_name.startswith("e_"):
        ini = None
        schedule = np.concatenate([[ini], schedule])
        drawstyle = "default"
    else:
        schedule = np.concatenate([schedule, schedule[-1:]])
        drawstyle = "steps-post"
    return schedule, drawstyle


def plot_entity(entity, schedule=None, ax=None, title=None):
    """
    Plot a single entity into axis.

    Parameters
    ----------
    entity : OptimizationEntity
        Entity that should be plotted.
    schedule : str or list, optional
       Schedule or list of schedules to save. Defaults to the current schedule.
       At most two schedules can be plotted.

       - `None` : Current schedule of entity
       - 'default' : Normal schedule
       - 'ref' : Reference schedule
    ax : matplotlib.Axes, optional
        Axes the Entity should be plotted into. Shows the plot in a new figure if not specified.
    title : str, optional
        Title of the plot. Uses the name of the entity if not specified.
    """
    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = None

    if schedule is None:
        schedule_names = [entity.current_schedule]
    elif isinstance(schedule, str):
        schedule_names = [schedule]
    elif isinstance(schedule, list):
        if len(schedule) > 2:
            raise ValueError("Amount of schedules to plot is too large.")
        schedule_names = schedule
    else:
        raise ValueError("Unknown type for schedule.")

    var_colors = _known_varnames.copy()
    unknown_vars = entity.schedule.keys() - _known_varnames.keys()
    if isinstance(entity, Battery):
        unknown_vars.discard("p_el_demand")
        unknown_vars.discard("p_el_supply")
        unknown_vars.discard("p_state")
    if isinstance(entity, ElectricVehicle):
        unknown_vars.discard("p_el_drive")
        unknown_vars.discard("p_state")
    if isinstance(entity, CombinedHeatPower):
        unknown_vars.discard("total_device")
        unknown_vars.discard("current_device")
    for i, uv in enumerate(unknown_vars):
        if i >= len(_other_colors):
            raise ValueError("Too many variables for entity to plot.")
        var_colors[uv] = _other_colors[i]
    linestyles = ['-', ':']
    min_y = np.infty
    max_y = -np.infty
    for i, entity_schedule in enumerate(schedule_names):
        for var_name, var_schedule in entity.schedules[entity_schedule].items():
            if var_name not in var_colors:
                continue
            c = var_colors[var_name][i]
            xs = [x * entity.time_slot for x in range(entity.simu_horizon+1)]
            extended_schedule, drawstyle = _extended_schedule(var_name, var_schedule)
            if schedule is None:
                label = var_name
            else:
                label = entity_schedule + "_" + var_name
            ax.plot(xs, extended_schedule, color=c, label=label, drawstyle=drawstyle, linewidth=2,
                    linestyle=linestyles[i])
            min_y = min(min_y, min(var_schedule))
            max_y = max(max_y, max(var_schedule))

    if min_y < 0.0:
        min_y = min_y*1.5
    elif min_y > 0.0:
        min_y = min_y * 0.5
    else:
        min_y = -0.1*max_y

    if max_y < 0.0:
        max_y = max_y*0.5
    elif max_y > 0.0:
        max_y = max_y * 1.5
    else:
        max_y = -0.1*min_y

    if title is None:
        ax.set_title(str(entity))
    else:
        ax.set_title(title)
    ax.set_ylim([min_y, max_y])
    ax.set_xlabel("Time in h")
    ax.legend(loc='upper right')
    if fig is not None:
        plt.show()
    return


def plot_imbalance(entity, schedule=None, var_name="p_el", ax=None, title=None):
    """
    Plot the imbalance of a schedule to its sub-entities.

    For entity containers and other similar entities with sub-entities the
    schedule of some variables should be the equal to the sum of the schedule
    of these variables of the sub-entities. This function plots this imbalance.

    Parameters
    ----------
    entity : OptimizationEntity
        Entity that should be plotted.
    schedule : str or list, optional
       Schedule or list of schedules to save. Defaults to the current schedule.
       At most two schedules can be plotted.

       - `None` : Current schedule of entity
       - 'default' : Normal schedule
       - 'ref' : Reference schedule
    var_name : str, optional
        The name of the variable to plot the imbalance for. Defaults to "p_el".
    ax : matplotlib.Axes, optional
        Axes the Entity should be plotted into. Shows the plot in a new figure if not specified.
    title : str, optional
        Title of the plot. Uses the name of the entity if not specified.
    """
    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = None

    if schedule is None:
        schedule_names = [entity.current_schedule]
    elif isinstance(schedule, str):
        schedule_names = [schedule]
    elif isinstance(schedule, list):
        schedule_names = schedule
    else:
        raise ValueError("Unknown type for schedule.")

    min_y = np.infty
    max_y = -np.infty
    for entity_schedule in schedule_names:
        imbalance = entity.schedules[entity_schedule][var_name].copy()
        for lower_entity in entity.get_lower_entities():
            imbalance -= lower_entity.schedules[entity_schedule][var_name]
        xs = [x * entity.time_slot for x in range(entity.simu_horizon + 1)]
        extended_schedule, drawstyle = _extended_schedule(var_name, imbalance)
        ax.plot(xs, extended_schedule, drawstyle=drawstyle, linewidth=1.5, label=entity_schedule)
        min_y = min(min_y, min(imbalance))
        max_y = max(max_y, max(imbalance))

    if min_y < 0.0:
        min_y = min_y*1.5
    elif min_y > 0.0:
        min_y = min_y * 0.5
    else:
        min_y = -0.1*max_y

    if max_y < 0.0:
        max_y = max_y*0.5
    elif max_y > 0.0:
        max_y = max_y * 1.5
    else:
        max_y = -0.1*min_y

    ax.set_ylim([min_y, max_y])
    if title is None:
        ax.set_title("Imbalance of entity {} for {}".format(str(entity), var_name))
    else:
        ax.set_title(title)
    ax.set_xlabel("Time in h")
    ax.legend(loc='upper right')
    if fig is not None:
        plt.show()
    return


def plot_entity_directory(entity, schedule=None, directory_path=None, levels=None, extension="png"):
    """
    Plot the entity and its sub-entities into a directory.

    Creates the directories and places the plots in them.
    Parameters
    ----------
    entity : OptimizationEntity
        Entity that should be plotted.
    schedule : str or list, optional
       Schedule or list of schedules to save. Defaults to the current schedule.
       At most two schedules can be plotted.
       `None` : Current schedule of entity
       'default' : Normal schedule
       'ref' : Reference schedule
    directory_path : str, optional
        Directory path in which plots are stored. Defaults to the name of the
        entity in the current directory.
    levels : int, optional
        The level of sub-entities to plot. Defaults to all sub-entities.

        - `0` : Only print city district.
        - `1` : Only print city district and buildings.
        - `2` : Print city district, buildings and their lower entities.
        - `...` : ...
    extension : str, optional
        File extension (i.e., the file format) of the figures to be stored in the given directory.
    """

    entities = [entity]
    directories = [""]
    if directory_path is None:
        directory_path = op.join(os.getcwd(), str(entity))
    os.makedirs(directory_path, exist_ok=True)
    while len(entities) != 0:
        next_entities = []
        next_directories = []
        for entity, directory in zip(entities, directories):
            if levels is None or levels > 0:
                lower_entities = list(entity.get_lower_entities())
            else:
                lower_entities = []
            if len(lower_entities) > 0 and directory != "":
                path = op.join(directory_path, directory, str(entity))
                os.makedirs(path, exist_ok=True)
            else:
                # first level entity or no lower entities
                path = op.join(directory_path, directory)

            fig, ax = plt.subplots()
            plot_entity(entity, schedule=schedule, ax=ax)
            fig.savefig(op.join(path, str(entity) + "." + extension))

            next_entities.extend(lower_entities)
            next_directories.extend([path] * len(lower_entities))

        if levels is not None:
            levels -= 1
        entities = next_entities
        directories = next_directories
    return



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
from pycity_scheduling.classes import *
from pycity_scheduling.algorithms import *


# This is a simple power scheduling example integrating grid constraints. The central optimization algorithm is used to
# schedule an alternative four load branch scenario compared to 'example_24_scheduling_with_grid_four_load_branch.py'.


def main(do_plot=False):
    print("\n\n------ Example 26: Scheduling with Grid - Four Load Branch Out ------\n\n")
    
    # Define timer, price, weather, and environment objects:
    nt = 24 * 7
    t = Timer(op_horizon=nt, step_size=3600)
    p = Prices(timer=t)
    w = Weather(timer=t)
    e = Environment(timer=t, weather=w, prices=p)
    # City district with district operator objective "price":
    cd = CityDistrict(environment=e)
    
    # Fixed load for buildings
    p_load = 0.03 * 1000  # kW
    
    # Building 1
    bd1 = Building(e, objective='price')
    cd.addEntity(entity=bd1, position=[0, 0])
    bes = BuildingEnergySystem(e)
    bd1.addEntity(bes)
    ap = Apartment(e)
    bd1.addEntity(ap)
    load = np.array([p_load] * nt)
    fi = FixedLoad(e, method=0, demand=load)
    ap.addEntity(fi)
    
    # Building 2
    bd2 = Building(e, objective='price')
    cd.addEntity(entity=bd2, position=[0, 0])
    bes = BuildingEnergySystem(e)
    bd2.addEntity(bes)
    ap = Apartment(e)
    bd2.addEntity(ap)
    load = np.array([p_load] * nt)
    fi = FixedLoad(e, method=0, demand=load)
    ap.addEntity(fi)
    
    # Building 3
    bd3 = Building(e, objective='price')
    cd.addEntity(entity=bd3, position=[0, 0])
    bes = BuildingEnergySystem(e)
    bd3.addEntity(bes)
    ap = Apartment(e)
    bd3.addEntity(ap)
    load = np.array([p_load] * nt)
    fi = FixedLoad(e, method=0, demand=load)
    ap.addEntity(fi)

    # Building 4
    bd4 = Building(e, objective='price')
    cd.addEntity(entity=bd4, position=[0, 0])
    bes = BuildingEnergySystem(e)
    bd4.addEntity(bes)
    ap = Apartment(e)
    bd4.addEntity(ap)
    load = np.array([p_load] * nt)
    fi = FixedLoad(e, method=0, demand=load)
    ap.addEntity(fi)
    
    # Define grid properties:
    P_ref = 1000  # kW
    Q_ref = 500  # kVar
    V_ref = 400  # V
    line_capacity = 3**0.5 * 0.242 * 400  # kW
    R_cable = 0.225  # Ohm/km
    X_cable = 0.08  # Ohm/km
    line_length = 0.05  # km
    
    cd.addElectricalGrid(e, P_ref, Q_ref, slack_V=0.998, ref_V=V_ref)

    # Add electrical grid:
    grid = cd.electrical_grid
    grid.update_cable_properties(R_cable, X_cable)

    # Add lines:
    bus3 = ElectricalNode(e, None, None)
    bus4 = ElectricalNode(e, None, None)
    bus5 = ElectricalNode(e, None, None)
    bus6 = ElectricalNode(e, None, None)

    grid.connect_entities(grid.slack_node, bus3, line_capacity, line_length)
    grid.connect_entities(bus3, bus4, line_capacity, line_length)
    grid.connect_entities(bus4, bus5, line_capacity, line_length)
    grid.connect_entities(bus5, bus6, line_capacity, line_length)

    grid.connect_entities(bus3, bd1, line_capacity, line_length)
    grid.connect_entities(bus4, bd2, line_capacity, line_length)
    grid.connect_entities(bus5, bd3, line_capacity, line_length)
    grid.connect_entities(bus6, bd4, line_capacity, line_length)

    # Perform the scheduling:
    opt = CentralOptimization(city_district=cd)
    results = opt.solve()
    cd.copy_schedule("central")
    
    # Print the building's schedules:
    print("Schedule building no. one:")
    print(list(bd1.p_el_schedule))
    print("Schedule building no. two:")
    print(list(bd2.p_el_schedule))
    print("Schedule building no. three:")
    print(list(bd2.p_el_schedule))
    print("Schedule building no. four:")
    print(list(bd2.p_el_schedule))
    
    # Print the district's schedule:
    print("Schedule of the city district:")
    print(list(cd.p_el_schedule))
    
    # Plot electrical grid's schedule:
    if do_plot:
        grid.plot_graph_results(t=1)
        

if __name__ == '__main__':
    # Run example:
    main(do_plot=True)



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np

import pycity_scheduling.util.factory as factory
from pycity_scheduling.util import calculate_flexibility_potential


# This example shows how to quantify the operational flexibility potential for a given city district using
# pycity_scheduling. It is built upon the same district setup as defined in example 'example_17_district_generator.py'.


def main(do_plot=False):
    print("\n\n------ Example 18: District Flexibility Quantification ------\n\n")

    # First, create an environment using the factory's "generate_standard_environment" method. The environment
    # automatically encapsulates time, weather, and price data/information.
    env = factory.generate_standard_environment(initial_date=(2018, 12, 6), step_size=900, op_horizon=96)

    # Create 12 single-family houses:
    num_sfh = 12

    # 50% SFH.2002, 30% SFH.2010, 20% SFH.2016 (based on TABULA):
    sfh_distribution = {
        'SFH.2002': 0.5,
        'SFH.2010': 0.3,
        'SFH.2016': 0.2,
    }

    # 50% of the single-family houses are equipped with heat pump, 10% with boiler, and 40% with electric heater:
    sfh_heating_distribution = {
        'HP': 0.5,
        'BL': 0.1,
        'EH': 0.4,
    }

    # All single-family houses are equipped with a fixed load, 20% have a deferrable load, and 30% have an electric
    # vehicle. Moreover, 50% of all single-family houses have a battery unit and 80% have a rooftop photovoltaic unit
    # installation.
    # The values are rounded in case they cannot be perfectly matched to the given number of buildings.
    sfh_device_probs = {
        'FL': 1.0,
        'DL': 0.2,
        'EV': 0.3,
        'BAT': 0.5,
        'PV': 0.8,
    }

    # Create 3 multi-family houses (number of apartments according to TABULA):
    num_mfh = 3

    # 60% MFH.2002, 20% SFH.2010, 20% SFH.2016 (based on TABULA):
    mfh_distribution = {
        'MFH.2002': 0.6,
        'MFH.2010': 0.2,
        'MFH.2016': 0.2,
    }

    # 40% of the multi-family houses are equipped with heat pump, 20% with boiler, and 40% with electric heater:
    mfh_heating_distribution = {
        'HP': 0.4,
        'BL': 0.2,
        'EH': 0.4,
    }

    # All apartments inside a multi-family houses are equipped with a fixed load, 20% have a deferrable load, and 20%
    # have an electric vehicle. Moreover, 40% of all multi-family houses have a battery unit and 80% have a rooftop
    # photovoltaic unit installation.
    # The values are rounded in case they cannot be perfectly matched to the given number of buildings.
    mfh_device_probs = {
        'FL': 1.0,
        'DL': 0.2,
        'EV': 0.2,
        'BAT': 0.4,
        'PV': 0.8,
    }

    # Finally, create the desired city district using the factory's "generate_tabula_district" method:
    district = factory.generate_tabula_district(environment=env,
                                                number_sfh=num_sfh,
                                                number_mfh=num_mfh,
                                                sfh_building_distribution=sfh_distribution,
                                                sfh_heating_distribution=sfh_heating_distribution,
                                                sfh_device_probabilities=sfh_device_probs,
                                                mfh_building_distribution=mfh_distribution,
                                                mfh_heating_distribution=mfh_heating_distribution,
                                                mfh_device_probabilities=mfh_device_probs)

    # The city district's maximum operational flexibility potential in terms of the absolute flexibility gain metric can
    # be quantified very easily using the pycity_scheduling's "calculate_flexibility_potential" function. The absolute
    # flexibility gain is a measure for the energy shifted between a coordinated and an uncoordinated scheduling
    # scenario. Here, the coordinated case is represented through the central optimization algorithm, whereas the
    # uncoordinated case is represented through the stand-alone optimization algorithm.
    flex = calculate_flexibility_potential(city_district=district,
                                           algorithm="central",
                                           reference_algorithm="stand-alone")

    # Print the flexibility metric result:
    np.set_printoptions(formatter={'float': '{: >8.3f}'.format})
    print('Absolute flex. gain (kWh):    {: >8.3f}'.format(flex))
    return


# Conclusions:
# In the given scenario, the absolute flexibility gain within the city district scenario is about 754kWh for the defined
# day (24h). This means that up to this volume of energy can be shifted because of the local flexibility potentials.
# For example, this flexibility potential can be used to achieve several system level services such as peak-shaving.
# The quantification of the flexibility potential is usually a crucial task when it comes to the initial assessment of
# local energy systems and/or city district energy systems.


if __name__ == '__main__':
    # Run example:
    main(do_plot=True)
